# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import os
import tempfile
import unittest
import warnings
import yaml

from ...provenance import MetadataInfo

from qiime2.core.util import md5sum


class YamlConstructorTests(unittest.TestCase):
    '''
    YAML Constructors are used to handle the custom YAML tags defined by the
    framework.
    '''
    def test_unknown_tag(self):
        '''
        Makes explicit the current handling of unimplemented custom tags. In
        future, we may want to deal with these more graciously (e.g. warn), but
        for now we're going to fail fast.
        '''
        tag = r"!foo 'this is not an implemented tag'"
        with self.assertRaisesRegex(
            yaml.constructor.ConstructorError,
            'could not determine a constructor.*!foo'
        ):
            yaml.safe_load(tag)

    def test_citation_key_constructor(self):
        tag = r"!cite 'framework|qiime2:2020.6.0.dev0|0'"
        actual = yaml.safe_load(tag)
        self.assertEqual(actual, 'framework|qiime2:2020.6.0.dev0|0')

    def test_color_primitive_constructor(self):
        tag = r"!color '#57f289'"
        actual = yaml.safe_load(tag)
        self.assertEqual(actual, '#57f289')

    def test_forward_ref_action_plugin_ref(self):
        tag = r"plugin: !ref 'environment:plugins:diversity'"
        actual = yaml.safe_load(tag)
        self.assertEqual(actual, {'plugin': 'diversity'})

    def test_forward_ref_generic_ref(self):
        tag = r"plugin: !ref 'environment:framework:version'"
        actual = yaml.safe_load(tag)
        exp = {'plugin': ['environment', 'framework', 'version']}
        self.assertEqual(exp, actual)

    def test_metadata_path_constructor(self):
        tag = r"!metadata 'metadata.tsv'"

        with tempfile.TemporaryDirectory() as tempdir:
            md_fp = os.path.join(tempdir, 'metadata.tsv')
            with open(md_fp, 'w') as fh:
                fh.write('rows and columns and stuff')

            action_fp = os.path.join(tempdir, 'action.yaml')
            with open(action_fp, 'w') as fh:
                fh.write(f'{tag}\n')

            with open(action_fp, 'r') as fh:
                actual = yaml.safe_load(fh)

            md5sum_hash = md5sum(md_fp)

        self.assertEqual(actual, MetadataInfo([], 'metadata.tsv', md5sum_hash))

    def test_metadata_path_constructor_one_Artifact_as_md(self):
        tag = r"!metadata '415409a4-stuff-e3eaba5301b4:feature_metadata.tsv'"

        with tempfile.TemporaryDirectory() as tempdir:
            md_fp = os.path.join(tempdir, 'feature_metadata.tsv')
            with open(md_fp, 'w') as fh:
                fh.write('rows and columns and stuff')

            action_fp = os.path.join(tempdir, 'action.yaml')
            with open(action_fp, 'w') as fh:
                fh.write(f'{tag}\n')

            with open(action_fp, 'r') as fh:
                actual = yaml.safe_load(fh)

            md5sum_hash = md5sum(md_fp)

        self.assertEqual(
            actual,
            MetadataInfo(
                ['415409a4-stuff-e3eaba5301b4'],
                'feature_metadata.tsv',
                md5sum_hash
            )
        )

    def test_metadata_path_constructor_many_Artifacts_as_md(self):
        tag = (
            r"!metadata '415409a4-stuff-e3eaba5301b4,12345-other-stuff-67890"
            r":feature_metadata.tsv'"
        )

        with tempfile.TemporaryDirectory() as tempdir:
            md_fp = os.path.join(tempdir, 'feature_metadata.tsv')
            with open(md_fp, 'w') as fh:
                fh.write('rows and columns and stuff')

            action_fp = os.path.join(tempdir, 'action.yaml')
            with open(action_fp, 'w') as fh:
                fh.write(f'{tag}\n')

            with open(action_fp, 'r') as fh:
                actual = yaml.safe_load(fh)

            md5sum_hash = md5sum(md_fp)

        self.assertEqual(
            actual,
            MetadataInfo(
                ['415409a4-stuff-e3eaba5301b4', '12345-other-stuff-67890'],
                'feature_metadata.tsv',
                md5sum_hash
            )
        )

    def test_no_provenance_constructor(self):
        tag = "!no-provenance '34b07e56-27a5-4f03-ae57-ff427b50aaa1'"
        with self.assertWarnsRegex(
            UserWarning, 'Artifact 34b07e.*prior to provenance'
        ):
            actual = yaml.safe_load(tag)
            self.assertEqual(actual, '34b07e56-27a5-4f03-ae57-ff427b50aaa1')

    def test_no_provenance_multiple_warnings_fire(self):
        tag_list = """
        - !no-provenance '34b07e56-27a5-4f03-ae57-ff427b50aaa1'
        - !no-provenance 'gerbil'
        """
        with warnings.catch_warnings(record=True) as w:
            # Just in case something else has modified the filter state
            warnings.simplefilter("default")
            yaml.safe_load(tag_list)
            # There should be exactly two warnings
            self.assertEqual(len(w), 2)

            # The first should be a Userwarning containing these strings
            self.assertEqual(UserWarning, w[0].category)
            self.assertIn('Artifact 34b07e', str(w[0].message))
            self.assertIn('prior to provenance', str(w[0].message))

            # And the second should look similar
            self.assertEqual(UserWarning, w[1].category)
            self.assertIn('gerbil', str(w[1].message))
            self.assertIn('prior to provenance', str(w[0].message))

    def test_set_ref(self):
        flow_tag = r"!set ['foo', 'bar', 'baz']"
        flow = yaml.safe_load(flow_tag)
        self.assertEqual(flow, {'foo', 'bar', 'baz'})

        # NOTE: we don't expect duplicate values here (because dumped values
        # were a set), but it doesn't hurt to test the behavior
        block_tag = '!set\n- spam\n- egg\n- spam\n'
        block = yaml.safe_load(block_tag)
        self.assertEqual(block, {'spam', 'egg'})



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import os
import pathlib
import shutil
import tempfile
import unittest
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Generator
import warnings
from zipfile import ZipFile, ZIP_DEFLATED

from ..parse import ProvDAG

import qiime2
from qiime2 import Artifact, Metadata, ResultCollection
from qiime2.core.archive import Archiver
from qiime2.sdk.plugin_manager import PluginManager


@dataclass
class DummyArtifact:
    name: str
    artifact: Artifact
    uuid: str
    filepath: str
    dag: ProvDAG
    archive_version: int = 6


class DummyArtifacts:
    def __init__(self):
        self.pm = PluginManager()
        self.dp = self.pm.plugins['dummy-plugin']
        self.tempdir = tempfile.mkdtemp(prefix='qiime2-dummy-artifacts-temp-')

        self.datadir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), 'data'
        )

        self.init_import_artifacts()
        self.init_action_artifacts()
        self.init_all_version_artifacts()
        self.init_artifact_with_md_in_provenance()
        self.init_no_checksum_dag()

    def init_import_artifacts(self):
        '''
        artifacts with only import in their provenance
        '''
        single_int = Artifact.import_data('SingleInt', 0)
        single_int2 = Artifact.import_data('SingleInt', 7)
        int_seq1 = Artifact.import_data('IntSequence1', [1, 1, 2])
        int_seq2 = Artifact.import_data('IntSequence2', [3, 5])
        mapping1 = Artifact.import_data('Mapping', {'a': 42})
        mapping2 = Artifact.import_data('Mapping', {'c': 8, 'd': 13})

        for name in (
            'single_int', 'single_int2', 'int_seq1', 'int_seq2', 'mapping1',
            'mapping2'
        ):
            artifact = locals()[name]
            fp = os.path.join(self.tempdir, f'{name}.qza')
            artifact.save(fp)
            test_artifact = DummyArtifact(
                name, artifact, str(artifact.uuid), fp, ProvDAG(fp)
            )
            setattr(self, name, test_artifact)

    def init_action_artifacts(self):
        '''
        artifacts that have at least one non-import action in their provenance
        '''
        concat_ints = self.dp.methods['concatenate_ints']
        split_ints = self.dp.methods['split_ints']
        merge_mappings = self.dp.methods['merge_mappings']
        identity_with_metadata = self.dp.actions['identity_with_metadata']
        identity_with_metadata_column = \
            self.dp.actions['identity_with_metadata_column']
        dict_of_ints = self.dp.actions['dict_of_ints']
        optional_artifacts_method = \
            self.dp.actions['optional_artifacts_method']

        concated_ints, = concat_ints(
            self.int_seq1.artifact, self.int_seq1.artifact,
            self.int_seq2.artifact, 7, 13
        )
        other_concated_ints, = concat_ints(
            self.int_seq1.artifact, self.int_seq1.artifact,
            self.int_seq2.artifact, 81, 64
        )
        splitted_ints, _ = split_ints(self.int_seq2.artifact)
        merged_mappings, = merge_mappings(
            self.mapping1.artifact, self.mapping2.artifact
        )

        # artifacts with input artifact viewed as metadata
        int_seq_with_md, = identity_with_metadata(
            self.int_seq1.artifact, self.mapping1.artifact.view(Metadata)
        )
        int_seq_with_md_column, = identity_with_metadata_column(
            self.int_seq1.artifact,
            self.mapping2.artifact.view(Metadata).get_column('c')
        )
        concated_ints_with_md_column, = concat_ints(
            self.int_seq1.artifact, int_seq_with_md_column,
            self.int_seq2.artifact, 69, 2001
        )

        # artifact with input collection
        ints_dict = {
            'int1': self.single_int.artifact,
            'int2': self.single_int2.artifact,
        }
        ints_collection = ResultCollection(ints_dict)
        ints_from_collection, = dict_of_ints(ints_collection)
        int_from_collection = ints_from_collection['int1']

        # artifact with optional inputs left to default None
        int_seq_optional_input, = optional_artifacts_method(
            self.int_seq1.artifact, 8
        )

        # artifact from pipeline
        typical_pipeline = self.dp.pipelines['typical_pipeline']
        _, _, _, pipeline_viz, _ = typical_pipeline(
            self.int_seq1.artifact, self.mapping1.artifact, False
        )

        for name in (
            'concated_ints', 'other_concated_ints', 'splitted_ints',
            'merged_mappings', 'pipeline_viz', 'int_seq_with_md',
            'concated_ints_with_md_column', 'int_from_collection',
            'int_seq_optional_input'
        ):
            artifact = locals()[name]
            if name == 'pipeline_viz':
                ext = '.qzv'
            else:
                ext = '.qza'

            fp = os.path.join(self.tempdir, f'{name}{ext}')
            artifact.save(fp)
            test_artifact = DummyArtifact(
                name, artifact, str(artifact.uuid), fp, ProvDAG(fp)
            )
            setattr(self, name, test_artifact)

    def init_all_version_artifacts(self):
        '''
        import artifacts for all archive versions (0-6), which are stored
        uncompressed in the test data directory--necessary because we can not
        make artifacts of non-current versions on the fly
        '''
        for version in range(0, 7):
            if version == 0:
                dirname = 'table-v0'
            else:
                dirname = f'concated-ints-v{version}'

            versioned_artifact_dir = os.path.join(self.datadir, dirname)
            temp_zf_path = os.path.join(self.tempdir, 'temp.zip')
            write_zip_file(temp_zf_path, versioned_artifact_dir)

            filename = f'{dirname}.qza'
            fp = os.path.join(self.tempdir, filename)

            if version == 0:
                shutil.copy(temp_zf_path, fp)
                a = None
            else:
                a = Artifact.load(temp_zf_path)
                a.save(fp)

            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=UserWarning)
                dag = ProvDAG(fp)

            assert len(dag.terminal_nodes) == 1
            terminal_node, *_ = dag.terminal_nodes
            uuid = terminal_node._uuid

            name = filename.replace('-', '_').replace('.qza', '')
            da = DummyArtifact(name, a, uuid, fp, dag, version)
            setattr(self, name, da)

    def init_artifact_with_md_in_provenance(self):
        dirname = 'concated-ints-with-md'
        artifact_dir = os.path.join(self.datadir, dirname)
        temp_zf_path = os.path.join(self.tempdir, 'temp.zip')
        write_zip_file(temp_zf_path, artifact_dir)
        filename = f'{dirname}.qza'
        fp = os.path.join(self.tempdir, filename)
        a = Artifact.load(temp_zf_path)
        a.save(fp)

        dag = ProvDAG(fp)
        terminal_node, *_ = dag.terminal_nodes
        uuid = terminal_node._uuid
        name = filename.replace('-', '_').replace('.qza', '')
        da = DummyArtifact(name, a, uuid, fp, dag, 6)
        setattr(self, name, da)

    def init_no_checksum_dag(self):
        '''
        create archive with missing checksums.md5
        '''
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=UserWarning)
            with generate_archive_with_file_removed(
                self.single_int.filepath,
                self.single_int.uuid,
                'checksums.md5'
            ) as altered_archive:
                self.dag_missing_md5 = ProvDAG(altered_archive)

    @property
    def all_artifact_versions(self):
        return (
            self.table_v0, self.concated_ints_v1, self.concated_ints_v2,
            self.concated_ints_v3, self.concated_ints_v4,
            self.concated_ints_v5, self.concated_ints_v6
        )

    def free(self):
        shutil.rmtree(self.tempdir)


def write_zip_file(zfp, unzipped_dir):
    zf = ZipFile(zfp, 'w', ZIP_DEFLATED)
    for root, dirs, files in os.walk(unzipped_dir):
        for file in files:
            filepath = os.path.join(root, file)
            zf.write(
                filepath, os.path.relpath(filepath, unzipped_dir)
            )
    zf.close()


class CustomAssertions(unittest.TestCase):
    def assertREAppearsOnlyOnce(self, text, only_once, msg=None):
        appears_once_re = \
            (f'(?s)^(?:(?!{only_once}).)*{only_once}(?!.*{only_once}).*$')
        self.assertRegex(text, appears_once_re, msg)


def is_root_provnode_data(fp):
    '''
    a filter predicate which returns metadata, action, citation,
    and VERSION fps with which we can construct a ProvNode
    '''
    # Handle provenance files...
    if 'provenance' in fp and 'artifacts' not in fp:
        if 'action.yaml' in fp or 'citations.bib' in fp:
            return True

    # then handle files available at root, which require a cast
    if pathlib.Path(fp).parts[1] in (
        'VERSION', 'metadata.yaml', 'checksums.md5'
    ):
        return True


@contextmanager
def generate_archive_with_file_removed(
    qzv_fp: str, root_uuid: str, file_to_drop: pathlib.Path
) -> Generator[pathlib.Path, None, None]:
    """
    Deleting files from zip archives is hard, so this makes a temporary
    copy of qzf_fp with fp_to_drop removed and returns a handle to this archive

    file_to_drop should represent the relative path to the file within the
    zip archive, excluding the root directory (named for the root UUID).

    e.g. `/d9e080bb-e245-4ab0-a2cf-0a89b63b8050/metadata.yaml` should be passed
    in as `metadata.yaml`

    adapted from https://stackoverflow.com/a/513889/9872253
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        tmp_arc = pathlib.Path(tmpdir) / 'mangled.qzv'
        fp_pfx = pathlib.Path(root_uuid)
        zin = ZipFile(qzv_fp, 'r')
        zout = ZipFile(str(tmp_arc), 'w')
        for item in zin.infolist():
            buffer = zin.read(item.filename)
            drop_filename = str(fp_pfx / file_to_drop)
            if (item.filename != drop_filename):
                zout.writestr(item, buffer)
        zout.close()
        zin.close()
        yield tmp_arc


@contextmanager
def monkeypatch_archive_version(patch_version):
    try:
        og_version = Archiver.CURRENT_FORMAT_VERSION
        Archiver.CURRENT_FORMAT_VERSION = patch_version
        yield
    finally:
        Archiver.CURRENT_FORMAT_VERSION = og_version


@contextmanager
def monkeypatch_framework_version(patch_version):
    try:
        og_version = qiime2.__version__
        qiime2.__version__ = patch_version
        yield
    finally:
        qiime2.__version__ = og_version


def write_zip_archive(zfp, unzipped_dir):
    with ZipFile(zfp, 'w') as zf:
        for root, dirs, files in os.walk(unzipped_dir):
            for file in files:
                path = os.path.join(root, file)
                archive_name = os.path.relpath(path, start=unzipped_dir)
                zf.write(path, arcname=archive_name)



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import os
import shutil
import tempfile
import unittest
import zipfile

import pytest

from qiime2 import Artifact
from qiime2.core.archive.archiver import ChecksumDiff
from qiime2.sdk.plugin_manager import PluginManager

from .._checksum_validator import validate_checksums, ValidationCode
from .testing_utilities import write_zip_archive


class ValidateChecksumTests(unittest.TestCase):
    def setUp(self):
        self.pm = PluginManager()
        self.dp = self.pm.plugins['dummy-plugin']
        self.tempdir = tempfile.mkdtemp(
            prefix='qiime2-test-checksum-validator-temp-'
        )

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_validate_checksums(self):
        int_seq = Artifact.import_data('IntSequence1', [1, 2, 3])
        fp = os.path.join(self.tempdir, 'int-seq.qza')
        int_seq.save(fp)

        with zipfile.ZipFile(fp) as zf:
            is_valid, diff = validate_checksums(zf)
        self.assertEqual(is_valid, ValidationCode.VALID)
        self.assertEqual(diff, ChecksumDiff({}, {}, {}))

    @pytest.mark.filterwarnings('ignore::UserWarning')
    def test_validate_checksums_invalid(self):
        '''
        Mangle an intact v5 Archive so that its checksums.md5 is invalid,
        and then confirm that we're catching all the changes we've made
        Specifically:
        - remove the root `<uuid>/metadata.yaml`
        - add a new file called '<uuid>/tamper.txt`
        - overwrite `<uuid>/provenance/citations.bib`
        '''
        int_seq = Artifact.import_data('IntSequence1', [1, 2, 3])
        fp = os.path.join(self.tempdir, 'int-seq-altered.qza')
        int_seq.save(fp)

        with tempfile.TemporaryDirectory() as tempdir:
            with zipfile.ZipFile(fp) as zf:
                zf.extractall(tempdir)

            uuid = os.listdir(tempdir)[0]
            root_dir = os.path.join(tempdir, uuid)
            os.remove(os.path.join(root_dir, 'metadata.yaml'))
            with open(os.path.join(root_dir, 'tamper.txt'), 'w') as fh:
                pass
            citations_path = \
                os.path.join(root_dir, 'provenance', 'citations.bib')
            with open(citations_path, 'w') as fh:
                fh.write('file overwritten\n')

            write_zip_archive(fp, tempdir)

        with zipfile.ZipFile(fp) as zf:
            is_valid, diff = validate_checksums(zf)

        self.assertEqual(is_valid, ValidationCode.INVALID)
        self.assertEqual(list(diff.added.keys()), ['tamper.txt'])
        self.assertEqual(list(diff.removed.keys()), ['metadata.yaml'])
        self.assertEqual(list(diff.changed.keys()),
                         ['provenance/citations.bib'])

    @pytest.mark.filterwarnings('ignore::UserWarning')
    def test_validate_checksums_checksums_missing(self):
        int_seq = Artifact.import_data('IntSequence1', [1, 2, 3])
        fp = os.path.join(self.tempdir, 'int-seq-missing-version.qza')
        int_seq.save(fp)

        with tempfile.TemporaryDirectory() as tempdir:
            with zipfile.ZipFile(fp) as zf:
                zf.extractall(tempdir)

            uuid = os.listdir(tempdir)[0]
            os.remove(os.path.join(tempdir, uuid, 'checksums.md5'))

            write_zip_archive(fp, tempdir)

        with zipfile.ZipFile(fp) as zf:
            is_valid, diff = validate_checksums(zf)

        self.assertEqual(is_valid, ValidationCode.INVALID)
        self.assertEqual(diff, None)



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from .context import Context
from .action import Action, Method, Visualizer, Pipeline
from .plugin_manager import PluginManager, UninitializedPluginManagerError
from .result import Result, Artifact, Visualization, ResultCollection
from .results import Results
from .util import parse_type, parse_format, type_from_ast
from ..core.cite import Citations
from ..core.exceptions import ValidationError, ImplementationError

__all__ = ['Result', 'Results', 'Artifact', 'Visualization',
           'ResultCollection', 'Action', 'Method', 'Visualizer', 'Pipeline',
           'PluginManager', 'parse_type', 'parse_format', 'type_from_ast',
           'Context', 'Citations', 'PARALLEL_CONFIG', 'ValidationError',
           'ImplementationError', 'UninitializedPluginManagerError']



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import bibtexparser as bp
from bibtexparser.bwriter import BibTexWriter
import networkx as nx
import os
import pathlib
import pkg_resources
import shutil
import tempfile
from uuid import uuid4
from dataclasses import dataclass, field
from typing import Dict, Iterator, List, Optional, Tuple, Union

from .archive_parser import ProvNode
from .parse import ProvDAG
from .usage_drivers import build_header, build_footer
from ..provenance import MetadataInfo

from qiime2.sdk import PluginManager
from qiime2.sdk.usage import Usage, UsageVariable
from qiime2.sdk.util import camel_to_snake


@dataclass
class ReplayConfig():
    '''
    Dataclass that stores various user-selected configuration options and
    other bits of information relevant to provenance replay.

    Parameters
    ----------
    use : Usage
        The usage driver to be used for provenance replay.
    dump_recorded_metadata : bool
        If True, replay should write the metadata recorded in provenance to
        disk in .tsv format.
    use_recorded_metadata : bool
        If True, replay should use the metadata recorded in provenance.
    pm : PluginManager
        The active instance of the QIIME 2 PluginManager.
    md_context_has_been_printed : bool
        A flag set by default and used internally, allows context to be
        printed once and only once.
    no_provenance_context_has_been_printed : bool
        Indicates whether the no-provenance context documentation has been
        printed.
    header : bool
        If True, an introductory how-to header should be rendered in the
        script.
    verbose : bool
        If True, progress is reported to stdout.
    md_out_dir : str
        The directory where caputred metadata should be written.
    '''
    use: Usage
    dump_recorded_metadata: bool = True
    use_recorded_metadata: bool = False
    pm: PluginManager = PluginManager()
    md_context_has_been_printed: bool = False
    no_provenance_context_has_been_printed: bool = False
    header: bool = True
    verbose: bool = False
    md_out_dir: str = ''


@dataclass
class ActionCollections():
    '''
    std_actions are all normal provenance-tracked q2 actions, arranged as:

    {
        <action_id>: {
            <output_node_uuid>: 'output_name',
            <output_node_2_uuid:> 'output_name_2'
        },
        <action_2_id> : ...
     }

    no_provenance_nodes can't be organized by action, and in some cases we
    don't know anything but UUID for them, so we can fit these in a list.
    '''
    std_actions: Dict[str, Dict[str, str]] = field(default_factory=dict)
    no_provenance_nodes: List[str] = field(default_factory=list)


@dataclass
class UsageVariableRecord:
    name: str
    variable: UsageVariable = None


@dataclass
class ResultCollectionRecord:
    collection_uuid: str
    members: Dict[str, str]


class ReplayNamespaces:
    '''
    A dataclass collection of objects that each track some useful bit of
    information relevant to replay/usage namespaces.

    Attributes
    ----------
    _usg_var_ns : dict
        The central usage variable namespace that ensures no namespace clashes.
        Maps artifact uuid to `UsageVariableRecord`.
    _action_ns : set of str
        A collection of unique action strings that look like
        `{plugin}_{action}_{sequential int}`.
    result_collection_ns : dict
        Used to keep track of result collection members during usage rendering.
        Structure is as follows:

        {
            action-id: {
                output-name: {
                    'collection_uuid': uuid,
                    'artifacts': {
                        uuid: key-in-collection,
                        (...),
                    }
                },
                (...),
            }
        }

        where the action-id and the output-name uniquely identify a result
        collection that came from some action, the `collection_uuid` key stores
        a uuid for the entire collection needed for querying the
        `usg_var_namespace`, and the `artifacts` key stores all result
        collection members along with their keys so they can be accessed
        properly.
    '''
    def __init__(self, dag=None):
        self._usg_var_ns = {}
        self._action_ns = set()
        if dag:
            self.result_collection_ns = \
                self.make_result_collection_namespace(dag)
            self.artifact_uuid_to_rc_uuid, self.rc_contents_to_rc_uuid = \
                self.make_result_collection_mappings()
        else:
            self.result_collection_ns = {}
            self.artifact_uuid_to_rc_uuid = {}
            self.rc_contents_to_rc_uuid = {}

    def add_usg_var_record(self, uuid, name, variable=None):
        '''
        Given a uuid, name, and optionally a usage variable, create a usage
        variable record and add it to the namespace.

        Parameters
        ----------
        uuid : str
            The uuid of the the artifact or result collection.
        name : str
            The not-yet-unique name of the artifact or result collection.
        variable : UsageVariable or None
            The optional UsageVariable instance to add to the record.

        Returns
        -------
        str
            The now-unique name of the artifact or result collection.
        '''
        unique_name = self._make_unique_name(name)

        self._usg_var_ns[uuid] = UsageVariableRecord(unique_name, variable)

        return unique_name

    def update_usg_var_record(self, uuid, variable):
        '''
        Given a uuid update the record to contain the passed usage variable.
        The record is assumed to already be present in the namespace.

        Parameters
        ----------
        uuid : str
            The uuid of the artifact or result collection for which to update
            the usage variable instance.
        variable : UsageVariable
            The usage variable to add to the record.
        '''
        self._usg_var_ns[uuid].variable = variable

    def get_usg_var_record(self, uuid):
        '''
        Given a uuid, return the corresponding usage variable record, or none
        if the uuid is not in the namespace.

        Parameters
        ----------
        uuid : str
            The uuid of the artifact or result collection for which to return
            the record.

        Returns
        -------
        UsageVariableRecord or None
            The record if the uuid was found, otherwise None.
        '''
        try:
            return self._usg_var_ns[uuid]
        except KeyError:
            return None

    def get_usg_var_uuid(self, name: str) -> str:
        '''
        Given a usage variable name, return its uuid, or raise KeyError if the
        name is not in the namespace.

        Parameters
        ----------
        name : str
            The name of the usage variable record of interest.

        Returns
        -------
        str
            The corresponding uuid of the record.

        Raises
        ------
        KeyError
            If the name is not found in the namespace.
        '''
        for uuid, record in self._usg_var_ns.items():
            if name == record.name:
                return uuid

        raise KeyError(
            f'The queried name \'{name}\' does not exist in the namespace.'
        )

    def _make_unique_name(self, name: str) -> str:
        '''
        Appends `_<some int>` to name, such that the returned name won't
        collide with any variable names that already exist in `usg_var_ns`.

        Parameters
        ----------
        name : str
            The variable name to make unique.

        Returns
        -------
        str
            The unique integer-appended variable name.
        '''
        counter = 0
        unique_name = f'{name}_{counter}'
        names = [record.name for record in self._usg_var_ns.values()]

        # no-provenance nodes are stored with angle brackets around them
        while unique_name in names or f'<{unique_name}>' in names:
            counter += 1
            unique_name = f'{name}_{counter}'

        return unique_name

    def make_result_collection_namespace(self, dag: nx.digraph) -> dict:
        '''
        Constructs the result collections namespaces from the parsed digraph.

        Parameters
        ----------
        dag : nx.digraph
            The digraph representing the parsed provenance.

        Returns
        -------
        dict
            The result collection namespace.
        '''
        rc_ns = {}
        for node in dag:
            provnode = dag.get_node_data(node)
            rc_key = provnode.action.result_collection_key
            if rc_key:
                # output result collection
                action_id = provnode.action.action_id
                output_name = provnode.action.output_name
                if action_id not in rc_ns:
                    rc_ns[action_id] = {}
                if output_name not in rc_ns[action_id]:
                    artifacts = {rc_key: provnode._uuid}
                    rc_ns[action_id][output_name] = ResultCollectionRecord(
                        collection_uuid=str(uuid4()), members=artifacts
                    )
                else:
                    rc_ns[action_id][output_name].members[rc_key] = \
                        provnode._uuid

        return rc_ns

    def make_result_collection_mappings(self) -> Tuple[Dict]:
        '''
        Builds two mappings:
            - one from artifact uuid to a tuple of the uuid of the result
              collection of which it is a member and its key in the collection
            - one from the hash of the result collection contents (both with
              and without keys) to the uuid of the result collection

        Returns
        -------
        tuple of dict
            The two result collection mappings.
        '''
        a_to_c = {}  # artifact uuid -> collection uuid
        c_to_c = {}  # hash of collection contents -> collection uuid
        for action_id in self.result_collection_ns:
            for output_name in self.result_collection_ns[action_id]:
                record = self.result_collection_ns[action_id][output_name]
                for key, uuid in record.members.items():
                    a_to_c[uuid] = (record.collection_uuid, key)

                hashed_contents = self.hash_result_collection(record.members)
                hashed_contents_with_keys = \
                    self.hash_result_collection_with_keys(record.members)

                c_to_c[hashed_contents] = record.collection_uuid
                c_to_c[hashed_contents_with_keys] = record.collection_uuid

        return a_to_c, c_to_c

    def hash_result_collection_with_keys(self, members: Dict) -> int:
        '''
        Hashes the contents of a result collection. Useful for finding
        corresponding usage variables when rendering the replay of result
        collections. Order of the input result collection is not taken into
        account (the result collections are ordered alphabetically by key).

        Parameters
        ----------
        members : dict
            The contents of a result collection, looks like:

            {
                'a': some-uuid,
                'b': some-other-uuid,
                (...)
            }

        Returns
        -------
        int
            The hashed contents.
        '''
        sorted_members = {key: members[key] for key in sorted(members)}
        hashable_members_with_keys = tuple(
            (key, value) for key, value in sorted_members.items()
        )

        return hash(hashable_members_with_keys)

    def hash_result_collection(self, members: Union[Dict, List]) -> int:
        '''
        Hashes a list of uuids. Useful for finding corresponding result
        collections that may have been cast to list of uuids. If a dict is
        input it is first converted to a list of values (uuids).

        Parameters
        ----------
        members : dict or list
            The contents of a result collection, either as a dict or list.

        Returns
        -------
        int
            The hashed contents.
        '''
        if type(members) is dict:
            members = list(members.values())

        sorted_members = list(sorted(members))
        hashable_members = tuple(uuid for uuid in sorted_members)

        return hash(hashable_members)

    def add_rc_member_to_ns(self, uuid, name, use):
        '''
        Accesses a result collection member of interest and adds it the
        central usage variable namespace.

        Parameters
        ----------
        uuid : str
            The uuid of the artifact of interest.
        name : str
            The desired name of the to-be-made usage variable.
        use : Usage
            The currently executing usage driver.
        '''
        collection_uuid, key = self.artifact_uuid_to_rc_uuid[uuid]
        collection_var = self.get_usg_var_record(collection_uuid).variable

        var_name = self.add_usg_var_record(uuid, name)

        usg_var = use.get_artifact_collection_member(
            var_name, collection_var, key
        )

        self.update_usg_var_record(uuid, usg_var)

    def uniquify_action_name(self, plugin: str, action: str) -> str:
        '''
        Creates a unique name by concatenating plugin, action, and a counter,
        and adds this name to _action_ns before returning it.

        Parameters
        ----------
        plugin : str
            The name of the plugin.
        action : str
            The name of the action.
        action_namespace : set of str
            The collection of unqiue action names.

        Returns
        -------
        str
            The unique action name.
        '''
        counter = 0
        plg_action_name = f'{plugin}_{action}_{counter}'
        while plg_action_name in self._action_ns:
            counter += 1
            plg_action_name = f'{plugin}_{action}_{counter}'
        self._action_ns.add(plg_action_name)

        return plg_action_name


def replay_provenance(
    usage_driver: Usage,
    payload: Union[str, ProvDAG],
    out_fp: str,
    validate_checksums: bool = True,
    parse_metadata: bool = True,
    recurse: bool = False,
    use_recorded_metadata: bool = False,
    suppress_header: bool = False,
    verbose: bool = False,
    dump_recorded_metadata: bool = True,
    md_out_dir: str = ''
):
    '''
    Renders usage examples describing a ProvDAG, producing an interface-
    specific executable.

    ProvDAG inputs retain their original config values. The
    `validate_checksums`, `parse_metadata`, `recurse`, and `verbose` parameters
    are disregarded if the payload is a ProvDAG.

    Parameters
    ----------
    usage_driver : Usage
        The type of Usage driver to be used. Currently intended to be either
        `ReplayPythonUsage` or `ReplayCLIUsage`.
    payload : str or ProvDAG
        A filepath to an artifact or directory containing artifacts, or the
        ProvDAG to be parsed.
    out_fp : str
        The filepath at which to write the rendered executable.
    validate_checksums : bool
        Whether to perform checksum validation on the input artifact.
    parse_metadata : bool
        Whether to parse study metadata recorded in provenance.
    recurse : bool
        Whether to recursively parse nested directories containing artifacts.
    use_recorded_metadata : bool
        Whether to use the metadata recorded in provenance.
    suppress_header : bool
        Whether to forgo rendering the header and footer that are included
        by default in replay scripts.
    verbose : bool
        Whether to print status messages during processing.
    dump_recorded_metadata : bool
        Whether to write the metadata recorded in provenance to disk.
    md_out_dir : str
        The directory in which to write the recorded metadata if desired.
    '''
    if type(payload) is ProvDAG:
        parse_metadata = payload.cfg.parse_study_metadata

    if not parse_metadata:
        if use_recorded_metadata:
            raise ValueError(
                'Metadata not parsed for replay. Re-run with parse_metadata, '
                'or set use_recorded_metadata to False.'
            )
        if dump_recorded_metadata:
            raise ValueError(
                'Metadata not parsed, so cannot be written to disk. Re-run '
                'with parse_metadata, or set dump_recorded_metadata to False.'
            )
        if md_out_dir:
            raise ValueError(
                'Metadata not parsed, so cannot be written to disk. Re-run '
                'with parse_metadata, or do not pass a metadata output '
                'filepath argument.'
            )

    if use_recorded_metadata and not dump_recorded_metadata:
        raise NotImplementedError(
            'In order to produce a replay script that uses metadata '
            'captured in provenance, that metadata must first be written to '
            'disk. Re-run with dump-recorded-metadata set to True, or '
            'use-recorded-metadata set to False.'
        )

    dag = ProvDAG(
        payload, validate_checksums, parse_metadata, recurse, verbose
    )
    cfg = ReplayConfig(
        use=usage_driver(),
        use_recorded_metadata=use_recorded_metadata,
        dump_recorded_metadata=dump_recorded_metadata,
        verbose=verbose, md_out_dir=md_out_dir
    )

    ns = ReplayNamespaces(dag)

    build_usage_examples(dag, cfg, ns)
    if not suppress_header:
        cfg.use.build_header()
        cfg.use.build_footer(dag)

    if cfg.dump_recorded_metadata:
        print('metadata written to recorded_metadata/')

    output = cfg.use.render(flush=True)
    with open(out_fp, mode='w') as out_fh:
        out_fh.write(output)


def build_usage_examples(
    dag: ProvDAG, cfg: ReplayConfig, ns: ReplayNamespaces
):
    '''
    Builds a chained usage example representing the analysis `dag`.

    Parameters
    ----------
    dag : ProvDAG
       The dag representation of parsed provenance.
    cfg : ReplayConfig
        Replay configuration options.
    ns : ReplayNamespaces
        Info tracking usage and result collection namespaces.
    '''
    sorted_nodes = nx.topological_sort(dag.collapsed_view)
    actions = group_by_action(dag, sorted_nodes, ns)

    for node_id in actions.no_provenance_nodes:
        node = dag.get_node_data(node_id)
        build_no_provenance_node_usage(node, node_id, ns, cfg)

    for action_id in (std_actions := actions.std_actions):
        # we are replaying actions not nodes, so any associated node works
        try:
            some_node_id = next(iter(std_actions[action_id]))
            node = dag.get_node_data(some_node_id)
        except KeyError:
            # we have result collection
            some_output_name = next(iter(ns.result_collection_ns[action_id]))
            some_node_id = next(iter(
                ns.result_collection_ns[action_id][
                    some_output_name].members.values()
            ))
            node = dag.get_node_data(some_node_id)

        if node.action.action_type == 'import':
            build_import_usage(node, ns, cfg)
        else:
            build_action_usage(node, ns, std_actions, action_id, cfg)


def group_by_action(
    dag: ProvDAG, nodes: Iterator[str], ns: ReplayNamespaces
) -> ActionCollections:
    '''
    This groups the nodes from a DAG by action, returning an ActionCollections
    aggregating the outputs related to each action.

    Takes an iterator of UUIDs, allowing us to influence the ordering of the
    grouping.

    In cases where a captured output_name is unavailable, we substitute the
    output data's Semantic Type, snake-cased because it will be used as
    a variable name if this data is rendered by ArtifactAPIUsage.

    Parameters
    ----------
    dag : ProvDAG
        The dag representation of parsed provenance.
    nodes : iterator of str
        An iterator over node uuids.
    ns : ReplayNamespaces
        Info tracking usage and result collection namespaces.

    Returns
    -------
    ActionCollections
        The outputs grouped by action.
    '''
    actions = ActionCollections()
    for node_id in nodes:
        if dag.node_has_provenance(node_id):
            node = dag.get_node_data(node_id)
            action_id = node.action.action_id
            output_name = node.action.output_name
            if output_name is None:
                output_name = camel_to_snake(node.type)

            if node.action.result_collection_key:
                rc_record = ns.result_collection_ns[action_id][output_name]
                node_id = rc_record.collection_uuid

            if action_id not in actions.std_actions:
                actions.std_actions[action_id] = {node_id: output_name}
            else:
                # for result collections we overwrite this but don't care
                actions.std_actions[action_id][node_id] = output_name
        else:
            actions.no_provenance_nodes.append(node_id)

    return actions


def build_no_provenance_node_usage(
    node: Optional[ProvNode],
    uuid: str,
    ns: ReplayNamespaces,
    cfg: ReplayConfig
):
    '''
    Given a ProvNode (with no provenance), make sure comments will be rendered
    explaining this, add an empty usage variable to the namespace and log the
    node. Returns nothing, modifying the passed usage instance in place.

    Parameters
    ----------
    node : ProvNode or None
        Either a no-provenance node, or None indicating that only `uuid`
        is available.
    uuid : str
        The uuid of the node/result.
    ns : ReplayNamespaces
        Info tracking usage and result collection namespaces.
    cfg : ReplayConfig
        Replay configuration options. Contains the modified usage driver.
    '''
    if not cfg.no_provenance_context_has_been_printed:
        cfg.no_provenance_context_has_been_printed = True
        cfg.use.comment(
            'One or more nodes have no provenance, so full replay is '
            'impossible. Any commands we were able to reconstruct have been '
            'rendered, with the string descriptions below replacing actual '
            'inputs.'
        )
        cfg.use.comment(
            'Original Node ID                       String Description'
        )
    if node is None:
        # the node is a !no-provenance input and we have only UUID
        var_name = 'no-provenance-node'
    else:
        var_name = camel_to_snake(node.type)

    ns.add_usg_var_record(uuid, var_name)

    # make a usage variable for downstream consumption
    empty_var = cfg.use.usage_variable(
        ns.get_usg_var_record(uuid).name, lambda: None, 'artifact'
    )
    ns.update_usg_var_record(uuid, empty_var)

    # log the no-prov node
    usg_var = ns.get_usg_var_record(uuid).variable
    cfg.use.comment(f"{uuid}   {usg_var.to_interface_name()}")


def build_import_usage(
    node: ProvNode, ns: ReplayNamespaces, cfg: ReplayConfig
):
    '''
    Given a ProvNode, adds an import usage example for it, roughly
    resembling the below. Returns nothing, modifying the passed usage instance
    in place.

    raw_seqs = use.init_format('raw_seqs', lambda: None, ext='fastq.gz')
    imported_seqs = use.import_from_format(
        'emp_single_end_sequences',
        'EMPSingleEndSequences',
        raw_seqs
    )
    The `lambda: None` is a placeholder for some actual data factory,
    and should not impact the rendered usage.

    Parameters
    ----------
    node : ProvNode
        The imported node of interest.
    ns : ReplayNamespaces
        Info tracking usage and result collection namespaces.
    cfg : ReplayConfig
        Replay configuration options. Contains the modified usage driver.
    '''
    format_id = node._uuid + '_f'
    ns.add_usg_var_record(format_id, camel_to_snake(node.type) + '_f')

    format_for_import = cfg.use.init_format(
        ns.get_usg_var_record(format_id).name, lambda: None
    )

    var_name = ns.add_usg_var_record(node._uuid, camel_to_snake(node.type))

    use_var = cfg.use.import_from_format(
        var_name, node.type, format_for_import
    )
    ns.update_usg_var_record(node._uuid, use_var)


def build_action_usage(
    node: ProvNode,
    ns: ReplayNamespaces,
    std_actions: Dict[str, Dict[str, str]],
    action_id: str,
    cfg: ReplayConfig
):
    '''
    Adds an action usage example to `use` for some ProvNode.
    Returns nothing, modifying the passed usage instance in place.

    use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='pielou_evenness'),
        use.UsageInputs(table=ft),
        use.UsageOutputNames(vector='pielou_vector')
    )

    Parameters
    ----------
    node : ProvNode
        The node the creating action of which is of interest.
    ns : ReplayNamespaces
        Info tracking usage and result collection namespaces.
    std_actions : dict
        Expalained in ActionCollections.
    action_id : str
        The uuid of the action.
    cfg : ReplayConfig
        Replay configuration options. Contains the modified usage driver.
    '''
    command_specific_md_context_has_been_printed = False
    plugin = node.action.plugin
    action = node.action.action_name
    plg_action_name = ns.uniquify_action_name(plugin, action)

    inputs = _collect_action_inputs(cfg.use, ns, node)

    # Process outputs before params so we can access the unique output name
    # from the namespace when dumping metadata to files below
    raw_outputs = std_actions[action_id].items()
    outputs = _uniquify_output_names(ns, raw_outputs)

    for param_name, param_val in node.action.parameters.items():
        # We can currently assume that None arguments are only passed to params
        # as default values, so we can skip these parameters entirely in replay
        if param_val is None:
            continue

        if isinstance(param_val, MetadataInfo):
            unique_md_id = ns.get_usg_var_record(node._uuid).name \
                           + '_' + param_name
            md_fn = ns.add_usg_var_record(
                unique_md_id, camel_to_snake(param_name)
            )
            if cfg.dump_recorded_metadata:
                md_with_ext = md_fn + '.tsv'
                dump_recorded_md_file(
                    cfg, node, plg_action_name, param_name, md_with_ext
                )

            if cfg.use_recorded_metadata:
                # the local dir and fp where md will be saved (if at all) is:
                md_fn = f'{plg_action_name}/{md_fn}'
                md = init_md_from_recorded_md(
                    node, param_name, unique_md_id, ns, cfg, md_fn
                )
            else:
                if not cfg.md_context_has_been_printed:
                    cfg.md_context_has_been_printed = True
                    cfg.use.comment(
                        "Replay attempts to represent metadata inputs "
                        "accurately, but metadata .tsv files are merged "
                        "automatically by some interfaces, rendering "
                        "distinctions between file inputs invisible in "
                        "provenance. We output the recorded metadata to disk "
                        "to enable visual inspection.")

                if not command_specific_md_context_has_been_printed:
                    if cfg.md_out_dir:
                        fp = f'{cfg.md_out_dir}/{plg_action_name}'
                    else:
                        fp = f'./recorded_metadata/{plg_action_name}/'

                    cfg.use.comment(
                        "The following command may have received additional "
                        "metadata .tsv files. To confirm you have covered "
                        "your metadata needs adequately, review the original "
                        f"metadata, saved at '{fp}'")

                if not param_val.input_artifact_uuids:
                    md = init_md_from_md_file(
                        node, param_name, unique_md_id, ns, cfg
                    )
                else:
                    md = init_md_from_artifacts(param_val, ns, cfg)

            param_val = md

        inputs.update({param_name: param_val})

    usg_var = cfg.use.action(
        cfg.use.UsageAction(plugin_id=plugin, action_id=action),
        cfg.use.UsageInputs(**inputs),
        cfg.use.UsageOutputNames(**outputs)
    )

    # add the usage variable(s) to the namespace
    for res in usg_var:
        uuid_key = ns.get_usg_var_uuid(res.name)
        ns.update_usg_var_record(uuid_key, res)


def _collect_action_inputs(
    use: Usage, ns: ReplayNamespaces, node: ProvNode
) -> dict:
    '''
    Returns a dict containing the action Inputs for a ProvNode.
    Dict structure: {input_name: input_var} or {input_name: [input_var1, ...]}.

    Parameters
    ----------
    use : Usage
        The currently executing usage driver.
    ns : ReplayNamespaces
        Info tracking usage and result collection namespaces.
    node : ProvNode
        The node the creating action of which's inputs are of interest.

    Returns
    -------
    dict
        Mapping input names to their corresponding usage variables.
    '''
    inputs_dict = {}
    for input_name, input_value in node.action.inputs.items():
        # Currently we can only have a None as a default value, so we can skip
        # this as it was not provided
        if input_value is None:
            continue

        # Received a single artifact
        if type(input_value) is str:
            if ns.get_usg_var_record(input_value) is None:
                ns.add_rc_member_to_ns(input_value, input_name, use)

            resolved_input = ns.get_usg_var_record(input_value).variable

        # Received a list of artifacts
        elif type(input_value) is list:
            # may be rc cast to list so search for equivalent rc
            # if not then follow algorithm for single str for each
            input_hash = ns.hash_result_collection(input_value)
            if collection_uuid := ns.rc_contents_to_rc_uuid.get(input_hash):
                # corresponding rc found
                resolved_input = ns.get_usg_var_record(
                    collection_uuid
                ).variable
            else:
                # find each artifact and assemble into a list
                input_list = []
                for input_value in input_value:
                    if ns.get_usg_var_record(input_value) is None:
                        ns.add_rc_member_to_ns(input_value, input_name, use)

                    input_list.append(
                        ns.get_usg_var_record(input_value).variable
                    )

                resolved_input = input_list

        # Received a dict of artifacts (ResultCollection)
        elif type(input_value) is dict:
            # search for equivalent rc if not found then create new rc
            rc = input_value
            input_hash = ns.hash_result_collection_with_keys(rc)
            if collection_uuid := ns.rc_contents_to_rc_uuid.get(input_hash):
                # corresponding rc found
                resolved_input = ns.get_usg_var_record(
                    collection_uuid
                ).variable
            else:
                # build new rc
                new_rc = {}
                for key, input_value in rc.items():
                    if ns.get_usg_var_record(input_value) is None:
                        ns.add_rc_member_to_ns(input_value, input_name, use)

                    new_rc[key] = ns.get_usg_var_record(input_value).variable

                # make new rc usg var
                new_collection_uuid = uuid4()
                var_name = ns.add_usg_var_record(
                    new_collection_uuid, input_name
                )
                usg_var = use.construct_artifact_collection(var_name, new_rc)
                ns.update_usg_var_record(new_collection_uuid, usg_var)
                resolved_input = ns.get_usg_var_record(
                    new_collection_uuid
                ).variable

        # If we ever mess with inputs again and add a new type here this should
        # trip otherwise we should never see it
        else:
            msg = f"Got a '{input_value}' as input which is of type" \
                  f" '{type(input_value)}'. Supported types are str, list," \
                  " and dict."
            raise ValueError(msg)

        inputs_dict[input_name] = resolved_input

    return inputs_dict


def _uniquify_output_names(
    ns: ReplayNamespaces, raw_outputs: dict
) -> dict:
    '''
    Returns a dict containing the uniquified output names from a ProvNode.
    Dict structure: {output_name: uniquified_output_name}.

    Parameters
    ----------
    ns : ReplayNamespaces
        Info tracking usage and result collection namespaces.
    raw_outputs : dict
        Mapping of node uuid to output-name as seen in action.yaml.

    Returns
    -------
    dict
        Mapping of original output-name to output-name after being made unique.
    '''
    outputs = {}
    for uuid, output_name in raw_outputs:
        var_name = ns.add_usg_var_record(uuid, output_name)
        outputs.update({output_name: var_name})

    return outputs


def init_md_from_recorded_md(
    node: ProvNode,
    param_name: str,
    md_id: str,
    ns: ReplayNamespaces,
    cfg: ReplayConfig,
    md_fn: str
) -> UsageVariable:
    '''
    Initializes and returns a Metadata UsageVariable from Metadata parsed
    from provenance.

    Parameters
    ----------
    node : ProvNode
        The node the creating action of which was passed metadata.
    param_name : str
        The name of the parameter to which metadata was passed.
    md_id : str
        Looks like: f'{node uuid}_{param name}'.
    ns : ReplayNamespaces
        Namespaces associated with provenance replay.
    cfg : ReplayConfig
        Replay configuration options. Contains the executing usage driver.
    md_fn : str
        Looks like: f'{plugin}_{action}_{counter}/{unique param name}.

    Returns
    -------
    UsageVariable
        Of type metadata or metadata column.

    Raises
    ------
    ValueError
        If the node has no metadata.
    '''
    if not node.metadata:
        raise ValueError(
            'This function should only be called if the node has metadata.'
        )

    md_df = node.metadata[param_name]

    def factory():
        from qiime2 import Metadata
        return Metadata(md_df)

    cwd = pathlib.Path.cwd()
    if cfg.md_out_dir:
        fn = str(cwd / cfg.md_out_dir / md_fn)
    else:
        fn = str(cwd / 'recorded_metadata' / md_fn)

    md = cfg.use.init_metadata(
        ns.get_usg_var_record(md_id).name, factory, dumped_md_fn=fn
    )
    plugin = node.action.plugin
    action = node.action.action_name

    if param_is_metadata_column(cfg, param_name, plugin, action):
        mdc_id = node._uuid + '_mdc'
        mdc_name = ns.get_usg_var_record(md_id).name + '_mdc'
        var_name = ns.add_usg_var_record(mdc_id, mdc_name)
        md = cfg.use.get_metadata_column(var_name, '<column name>', md)

    return md


def init_md_from_md_file(
    node: ProvNode,
    param_name: str,
    md_id: str,
    ns: ReplayNamespaces,
    cfg: ReplayConfig
) -> UsageVariable:
    '''
    Initializes and returns a Metadata UsageVariable with no real data,
    mimicking a user passing md as a .tsv file.

    Parameters
    ----------
    node : ProvNode
        The node the creating action of which was passed a metadata file.
    param_name : str
        The parameter name to which the metadata file was passed.
    md_id : str
        Looks like: f'{node uuid}_{param name}'.
    ns : ReplayNamespaces
        Namespaces associated with provenance replay.
    cfg : ReplayConfig
        Replay configuration options. Contains the executing usage driver.

    Returns
    -------
    UsageVariable
        Of type metadata or metadata column.
    '''
    plugin = node.action.plugin
    action = node.action.action_name
    md = cfg.use.init_metadata(ns.get_usg_var_record(md_id).name, lambda: None)

    if param_is_metadata_column(cfg, param_name, plugin, action):
        mdc_id = node._uuid + '_mdc'
        mdc_name = ns.get_usg_var_record(md_id).name + '_mdc'
        var_name = ns.add_usg_var_record(mdc_id, mdc_name)
        md = cfg.use.get_metadata_column(var_name, '<column name>', md)

    return md


def init_md_from_artifacts(
        md_inf: MetadataInfo, ns: ReplayNamespaces, cfg: ReplayConfig
) -> UsageVariable:
    '''
    Initializes and returns a Metadata UsageVariable with no real data,
    mimicking a user passing one or more QIIME 2 Artifacts as metadata.

    We expect these usage vars are already in the namespace as artifacts if
    we're reading them in as metadata.

    Parameters
    ----------
    md_inf : MetadataInfo
        Named tuple with fields `input_artifact_uuids` which is a list of
        uuids and `relative_fp` which is the filename of the metadata file.
        These are parsed from a !metadata tag in action.yaml.
    ns : ReplayNamespaces
        Info tracking usage and result collection namespaces.
    cfg: ReplayConfig
        Replay configuration options. Contains the executing usage driver.

    Returns
    -------
    UsageVariable
        Of type metadata.

    Raises
    ------
    ValueError
        If no input artifact uuids are present in MetadataInfo.
    '''
    if not md_inf.input_artifact_uuids:
        raise ValueError(
            'This funtion should not be used if '
            'MetadataInfo.input_artifact_uuids is empty.'
        )

    md_files_in = []
    for artifact_uuid in md_inf.input_artifact_uuids:
        amd_id = artifact_uuid + '_a'
        var_name = ns.get_usg_var_record(artifact_uuid).variable.name + '_a'
        if ns.get_usg_var_record(amd_id) is None:
            var_name = ns.add_usg_var_record(amd_id, var_name)
            art_as_md = cfg.use.view_as_metadata(
                var_name, ns.get_usg_var_record(artifact_uuid).variable
            )
            ns.update_usg_var_record(amd_id, art_as_md)
        else:
            art_as_md = ns.get_usg_var_record(amd_id).variable

        md_files_in.append(art_as_md)

    if len(md_inf.input_artifact_uuids) > 1:
        # we can't uniquify this normally, because one uuid can be merged with
        # combinations of others
        merge_id = '-'.join(md_inf.input_artifact_uuids)
        var_name = ns.add_usg_var_record(merge_id, 'merged_artifacts')
        merged_md = cfg.use.merge_metadata(
            var_name, *md_files_in
        )
        ns.update_usg_var_record(merge_id, merged_md)

    return art_as_md


def dump_recorded_md_file(
    cfg: ReplayConfig,
    node: ProvNode,
    action_name: str,
    md_id: str,
    fn: str
):
    '''
    Writes one metadata DataFrame pointed to by `md_id` to a .tsv file.
    Each action gets its own directory containing relevant md files.

    Raises a ValueError if the node has no metadata

    Parameters
    ----------
    cfg : ReplayConfig
        Replay configuration options. Contains the executing usage driver.
    node : ProvNode
        The node the creating action of which recorded metadata as input. Used
        here only to ensure that metadata was in fact recorded.
    action_name : str
        Looks like: f'{plugin}_{action}_{counter}'.
    md_id : str
        Looks like: f'{node uuid}_{param name}'.
    fn : str
        Looks like: f'{unique param name}.tsv'.

    Raises
    ------
    ValueError
        If the passed node does not have metadata in its creating action.
    '''
    if node.metadata is None:
        raise ValueError(
            'This function should only be called if the node has metadata.'
        )

    if cfg.md_out_dir:
        md_out_dir_base = pathlib.Path(cfg.md_out_dir)
    else:
        cwd = pathlib.Path.cwd()
        md_out_dir_base = cwd / 'recorded_metadata'

    action_dir = md_out_dir_base / action_name
    action_dir.mkdir(parents=True, exist_ok=True)

    md_df = node.metadata[md_id]
    out_fp = action_dir / (fn)
    md_df.to_csv(out_fp, sep='\t', index=False)


def param_is_metadata_column(
    cfg: ReplayConfig, param: str, plugin: str, action: str
) -> bool:
    '''
    Returns True if the parameter name `param` is registered as a
    MetadataColumn.

    Parameters
    ----------
    cfg : ReplayConfig
        Replay configuration options. Contains the plugin manager object.
    param : str
        The name of the parameter of interest.
    plugin : str
        The plugin that the relevant action belongs to.
    action : str
        The action that has the parameter of interest.

    Returns
    -------
    bool
        Indicating whether the parameter of interest is a MetadataColumn.

    Raises
    ------
    KeyError
        - If the plugin of interest is not registered with the plugin manager.
        - If the action of interest is not registered with the plugin.
        - If the parameter is not in the signature of the action.
    '''
    plugin = cfg.pm.get_plugin(id=plugin)

    try:
        action_f = plugin.actions[action]
    except KeyError:
        raise KeyError(
            f'No action registered with name {action} in plugin {plugin}.'
        )

    try:
        param_spec = action_f.signature.parameters[param]
    except KeyError:
        raise KeyError(
            f'No parameter registered with name {param} in action {action}.'
        )

    # HACK, but it works without relying on Q2's type system
    return 'MetadataColumn' in str(param_spec.qiime_type)


def collect_citations(
    dag: ProvDAG, deduplicate: bool = True
) -> bp.bibdatabase.BibDatabase:
    '''
    Returns a BibDatabase of all unique citations from a ProvDAG.
    If `deduplicate` is True references will be heuristically deduplicated.

    Parameters
    ----------
    dag : ProvDAG
        The ProvDAG object whose nodes contain citations to collect.
    deduplicate : bool
        Whether to deduplicate redundant citations.

    Returns
    -------
    bp.bibdatabase.BibDatabase
        A BibDatabase object containing the collected citations in bibtex
        format.
    '''
    bdb = bp.bibdatabase.BibDatabase()
    citations = []
    for node_uuid in dag:
        node = dag.get_node_data(node_uuid)
        # Skip no-prov nodes, which never have citations anyway
        if node is not None:
            node_citations = list(node.citations.values())
            citations.extend(node_citations)
    if deduplicate:
        citations = dedupe_citations(citations)

    bdb.entries = citations
    return bdb


class BibContent():
    '''
    A hashable data container capturing common bibtex fields

    Has many fields because keeping true duplicates is preferable to
    deduplicating two true non-duplicates.

    Parameters
    ----------
    entry : dict
        A dictionary of bibtex entries.
    '''
    def __init__(self, entry):
        self.title = entry.get('title'),
        self.author = entry.get('author'),
        self.journal = entry.get('journal')
        self.booktitle = entry.get('booktitle')
        self.year = entry.get('year')
        self.pages = entry.get('pages')

    def __eq__(self, other):
        return (
            type(self) is type(other) and
            self.title == other.title and
            self.author == other.author and
            self.journal == other.journal and
            self.booktitle == other.booktitle and
            self.year == other.year and
            self.pages == other.pages
        )

    def __hash__(self):
        return hash(
            str(self.title)
            + str(self.author)
            + str(self.journal)
            + str(self.journal)
            + str(self.booktitle)
            + str(self.year)
            + str(self.pages)
        )


def dedupe_citations(citations: List[Dict]) -> List[Dict]:
    '''
    Deduplicates citations based on bibtex id, bibtex content, and DOI.
    Citations are not guaranteed to be truly unique after deduplicating based
    on these values.

    Ensures only one qiime2 framework citation.

    Parameters
    ----------
    citations : list of dict
        The possibly redundant citations, each dict is a bibtex citation.

    Returns
    -------
    list of dict
        The deduplicated citations.
    '''
    deduped_citations = []
    is_framework_cited = False
    id_set = set()
    doi_set = set()
    content_set = set()
    for entry in citations:
        citation_id = entry['ID']

        if 'framework|qiime2' in citation_id:
            if not is_framework_cited:
                root = pkg_resources.resource_filename('qiime2', '.')
                root = os.path.abspath(root)
                path = os.path.join(root, 'citations.bib')
                with open(path) as bibtex_file:
                    q2_entry = bp.load(bibtex_file).entries.pop()

                q2_entry['ID'] = citation_id
                id_set.add(citation_id)
                deduped_citations.append(q2_entry)
                is_framework_cited = True
            continue

        # dedupe on id
        if citation_id in id_set:
            continue

        # dedupe on content
        entry_content = BibContent(entry)
        if entry_content in content_set:
            continue
        else:
            content_set.add(entry_content)

        # dedupe on doi if present
        doi = entry.get('doi')
        if doi is None:
            id_set.add(citation_id)
            deduped_citations.append(entry)
        elif doi not in doi_set:
            id_set.add(citation_id)
            doi_set.add(doi)
            deduped_citations.append(entry)

    return deduped_citations


def replay_citations(
    dag: ProvDAG,
    out_fp: str,
    deduplicate: bool = True,
    suppress_header: bool = False
):
    '''
    Writes a bibtex file containing all citations from a ProvDAG to disk.
    If `deduplicate` is True citations will be deduplicated, see
    `dedupe_citations()` for details.

    Parameters
    ----------
    dag : ProvDAG
        The provenance graph from which to collect citations.
    out_fp : str
        The filepath to which to write the citations.
    deduplicate : bool
        Whether to deduplicate the collected citations.
    suppress_header : bool
        Whether to forgo adding a header and footer to the output file.
    '''
    bib_db = collect_citations(dag, deduplicate=deduplicate)
    boundary = '#' * 79
    header = []
    footer = []
    extra = [
        '',
        '# This bibtex-formatted citation file can be imported into '
        'popular citation ',
        '# managers like Zotero and Mendeley, simplifying management and '
        'formatting.'
    ]
    if not suppress_header:
        header = build_header(boundary=boundary, extra_text=extra) + ['\n']
        footer = build_footer(dag=dag, boundary=boundary)
    if bib_db.entries_dict == {}:
        bib_db = 'No citations were registered to the used Actions.'
        with open(out_fp, 'w') as bibfile:
            bibfile.write(bib_db)
    else:
        with open(out_fp, 'w') as bibfile:
            bibfile.write('\n'.join(header))
            bibfile.write(BibTexWriter().write(bib_db))
            bibfile.write('\n'.join(footer))


def replay_supplement(
    usage_drivers: List[Usage],
    payload: Union[str, ProvDAG],
    out_fp: str,
    validate_checksums: bool = True,
    parse_metadata: bool = True,
    use_recorded_metadata: bool = False,
    recurse: bool = False,
    deduplicate: bool = True,
    suppress_header: bool = False,
    verbose: bool = True,
    dump_recorded_metadata: bool = True
):
    '''
    Produces a zipfile package of useful documentation for in silico
    reproducibility of some QIIME 2 Result(s) from a ProvDAG, a QIIME 2
    Artifact, or a directory of Artifacts.

    Package includes:
    - replay scripts for all supported interfaces
    - a bibtex-formatted collection of all citations

    ProvDAG inputs retain their original config values. The
    `validate_checksums`, `parse_metadata`, `recurse`, and `verbose` parameters
    are disregarded if the payload is a ProvDAG.

    Parameters
    ----------
    usage_drivers : list of Usage
        The types of Usage drivers to use. Currently intended to consist of
        `ReplayPythonUsage`, `ReplayCLIUsage`.
    payload : str or ProvDAG
        A filepath to an artifact or directory containing artifacts, or the
        ProvDAG to be parsed.
    out_fp : str
        The filepath to which to write the zip file.
    validate_checksums : bool
        Whether to perform checksum validation on the input artifact.
    parse_metadata : bool
        Whether to parse study metadata recorded in provenance.
    use_recorded_metadata : bool
        Whether to use the metadata recorded in provenance.
    recurse : bool
        Whether to recursively parse nested directories containing artifacts.
    deduplicate : bool
        Whether to deduplicate citations collected from provenance.
    suppress_header : bool
        Whether to forgo rendering the header and footer that are included
        by default in replay scripts.
    verbose : bool
        Whether to print status messages during processing.
    dump_recorded_metadata : bool
        Whether to write the metadata recorded in provenance to disk.
    '''
    dag = ProvDAG(
        artifact_data=payload,
        validate_checksums=validate_checksums,
        parse_metadata=parse_metadata,
        recurse=recurse,
        verbose=verbose
    )
    with tempfile.TemporaryDirectory() as tempdir:
        tempdir_path = pathlib.Path(tempdir)
        arc_root = tempdir_path / pathlib.Path(out_fp).stem
        os.makedirs(arc_root)

        drivers_to_filenames = {
            'ReplayPythonUsage': 'python3_replay.py',
            'ReplayCLIUsage': 'cli_replay.sh',
        }

        for usage_driver in usage_drivers:
            if usage_driver.__name__ not in drivers_to_filenames:
                continue

            rel_fp = drivers_to_filenames[usage_driver.__name__]
            md_out_dir = arc_root / 'recorded_metadata'
            tmp_fp = arc_root / rel_fp
            replay_provenance(
                usage_driver=usage_driver,
                payload=dag,
                out_fp=str(tmp_fp),
                use_recorded_metadata=use_recorded_metadata,
                suppress_header=suppress_header,
                verbose=verbose,
                dump_recorded_metadata=dump_recorded_metadata,
                md_out_dir=md_out_dir
            )
            print(
                f'The {usage_driver.__name__} replay script was written to '
                f'{rel_fp}.'
            )

        citations_fp = arc_root / 'citations.bib'
        replay_citations(
            dag,
            out_fp=str(citations_fp),
            deduplicate=deduplicate,
            suppress_header=suppress_header
        )
        print('The citations bibtex file was written to citations.bib.')

        out_fp = pathlib.Path(os.path.realpath(out_fp))
        if out_fp.suffix == '.zip':
            out_fp = out_fp.with_suffix('')

        shutil.make_archive(out_fp, 'zip', tempdir)
        print(f'The reproducibility package was written to {out_fp}.zip.')



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
"""
Software to support scientific reproducibility, attribution, and
collaboration on the QIIME 2 platform.

Core objects:
- ProvDAG: A directed, acyclic graph (DAG) describing QIIME 2 provenance
- ProvNode: Parsed data about a single QIIME 2 Result, generated internally by
  a provenance parser and available through the ProvDAG
"""

from .parse import ProvDAG, archive_not_parsed
from .replay import (
    replay_provenance, replay_citations, replay_supplement,
)
from .util import get_root_uuid, get_nonroot_uuid
from .usage_drivers import ReplayPythonUsage
from .tests.testing_utilities import DummyArtifacts

__all__ = [
    'ProvDAG', 'archive_not_parsed', 'get_root_uuid', 'get_nonroot_uuid',
    'replay_provenance', 'replay_citations', 'replay_supplement',
    'ReplayPythonUsage', 'DummyArtifacts'
]



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2022, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import tempfile
import unittest
import pkg_resources

import parsl
from parsl.executors.threads import ThreadPoolExecutor
from parsl.errors import NoDataFlowKernelError

from qiime2 import Artifact, Cache
from qiime2.core.util import load_action_yaml
from qiime2.core.testing.type import SingleInt
from qiime2.core.testing.util import get_dummy_plugin
from qiime2.sdk.parallel_config import (PARALLEL_CONFIG, _TEST_EXECUTOR_,
                                        _MASK_CONDA_ENV_, ParallelConfig,
                                        get_config_from_file)


class TestConfig(unittest.TestCase):
    # Get actions
    plugin = get_dummy_plugin()
    pipeline = plugin.pipelines['resumable_pipeline']
    method = plugin.methods['list_of_ints']

    # Expected provenance based on type of executor used
    tpool_expected = [{
        'type': 'parsl', 'parsl_type': 'ThreadPoolExecutor'}, {
        'type': 'parsl', 'parsl_type': 'ThreadPoolExecutor'}]
    test_expected = [{
        'type': 'parsl', 'parsl_type': '_TEST_EXECUTOR_'}, {
        'type': 'parsl', 'parsl_type': '_TEST_EXECUTOR_'}]

    def setUp(self):
        # Create config
        self.test_default = parsl.Config(
            executors=[
                ThreadPoolExecutor(
                    max_threads=1,
                    label='tpool'
                ),
                _TEST_EXECUTOR_(
                    max_threads=1,
                    label='default'
                )
            ],
            # AdHoc Clusters should not be setup with scaling strategy.
            strategy='none',
        )

        # Create temp test dir and cache in dir
        self.test_dir = tempfile.TemporaryDirectory(prefix='qiime2-test-temp-')
        self.cache = Cache(os.path.join(self.test_dir.name, 'new_cache'))

        # Create artifacts here so we have unique inputs in each test
        self.art = [Artifact.import_data(SingleInt, 0),
                    Artifact.import_data(SingleInt, 1)]

        # Get paths to config files
        self.config_fp = self.get_data_path('test_config.toml')
        self.mapping_config_fp = self.get_data_path('mapping_config.toml')
        self.mapping_only_config_fp = \
            self.get_data_path('mapping_only_config.toml')
        self.complex_config_fp = \
            self.get_data_path('complex_config.toml')

    def tearDown(self):
        self.test_dir.cleanup()

        # Ensure default state post test
        PARALLEL_CONFIG.parallel_config = None
        PARALLEL_CONFIG.action_executor_mapping = {}

    def get_data_path(self, filename):
        return pkg_resources.resource_filename('qiime2.sdk.tests',
                                               'data/%s' % filename)

    def test_default_config(self):
        with ParallelConfig():
            self.assertIsInstance(
                PARALLEL_CONFIG.parallel_config, parsl.Config)
            self.assertEqual(PARALLEL_CONFIG.action_executor_mapping, {})

    def test_mapping_from_config(self):
        config, mapping = get_config_from_file(self.mapping_config_fp)

        with self.cache:
            with ParallelConfig(config, mapping):
                future = self.pipeline.parallel(self.art, self.art)
                list_return, dict_return = future._result()

        list_execution_contexts = self._load_alias_execution_contexts(
            list_return)
        dict_execution_contexts = self._load_alias_execution_contexts(
            dict_return)

        self.assertEqual(list_execution_contexts, self.test_expected)
        self.assertEqual(dict_execution_contexts, self.tpool_expected)

    def test_mapping_only_config(self):
        _, mapping = get_config_from_file(self.mapping_only_config_fp)

        with self.cache:
            with ParallelConfig(action_executor_mapping=mapping):
                future = self.pipeline.parallel(self.art, self.art)
                list_return, dict_return = future._result()

        list_execution_contexts = self._load_alias_execution_contexts(
            list_return)
        dict_execution_contexts = self._load_alias_execution_contexts(
            dict_return)

        self.assertEqual(list_execution_contexts, self.test_expected)
        self.assertEqual(dict_execution_contexts, self.tpool_expected)

    def test_mapping_from_dict(self):
        mapping = {'list_of_ints': 'test'}

        with self.cache:
            with ParallelConfig(action_executor_mapping=mapping):
                future = self.pipeline.parallel(self.art, self.art)
                list_return, dict_return = future._result()

        list_execution_contexts = self._load_alias_execution_contexts(
            list_return)
        dict_execution_contexts = self._load_alias_execution_contexts(
            dict_return)

        self.assertEqual(list_execution_contexts, self.test_expected)
        self.assertEqual(dict_execution_contexts, self.tpool_expected)

    def test_parallel_configs(self):
        with self.cache:
            with ParallelConfig():
                future = self.pipeline.parallel(self.art, self.art)
                list_return, dict_return = future._result()

            list_execution_contexts = self._load_alias_execution_contexts(
                list_return)
            dict_execution_contexts = self._load_alias_execution_contexts(
                dict_return)

            self.assertEqual(list_execution_contexts, self.tpool_expected)
            self.assertEqual(dict_execution_contexts, self.tpool_expected)

            with ParallelConfig(self.test_default):
                future = self.pipeline.parallel(self.art, self.art)
                list_return, dict_return = future._result()

            list_execution_contexts = self._load_alias_execution_contexts(
                list_return)
            dict_execution_contexts = self._load_alias_execution_contexts(
                dict_return)

            self.assertEqual(list_execution_contexts, self.test_expected)
            self.assertEqual(dict_execution_contexts, self.test_expected)

            # At this point we should be using the default config again which
            # does not have an executor called tpool
            with ParallelConfig(
                    action_executor_mapping={'list_of_ints': 'tpool'}):
                with self.assertRaisesRegex(KeyError, 'tpool'):
                    future = self.pipeline.parallel(self.art, self.art)
                    list_return, dict_return = future._result()

    def test_nested_configs(self):
        with self.cache:
            with self.assertRaisesRegex(
                    ValueError, 'cannot nest ParallelConfigs'):
                with ParallelConfig():
                    with ParallelConfig(self.test_default):
                        pass

    def test_parallel_non_pipeline(self):
        with self.assertRaisesRegex(
                ValueError, 'Only pipelines may be run in parallel'):
            self.method.parallel(self.art)

    def test_no_vendored_fp(self):
        with _MASK_CONDA_ENV_():
            with ParallelConfig():
                with self.cache:
                    future = self.pipeline.parallel(self.art, self.art)
                    list_return, dict_return = future._result()

            list_execution_contexts = self._load_alias_execution_contexts(
                list_return)
            dict_execution_contexts = self._load_alias_execution_contexts(
                dict_return)

            self.assertEqual(list_execution_contexts, self.tpool_expected)
            self.assertEqual(dict_execution_contexts, self.tpool_expected)

    def test_load_complex_config(self):
        """ Test that all parsl modules we currently map are correct
        """
        config, mapping = get_config_from_file(self.complex_config_fp)
        # Just assert that we were able to parse the file and get a config out
        with ParallelConfig(config, mapping):
            self.assertIsInstance(
                PARALLEL_CONFIG.parallel_config, parsl.Config)
            self.assertEqual(PARALLEL_CONFIG.action_executor_mapping, {})

    def test_no_config(self):
        with self.assertRaisesRegex(NoDataFlowKernelError,
                                    'Must first load config'):
            self.pipeline.parallel(self.art, self.art)

    def test_config_unset(self):
        with ParallelConfig():
            self.pipeline.parallel(self.art, self.art)

        with self.assertRaisesRegex(NoDataFlowKernelError,
                                    'Must first load config'):
            self.pipeline.parallel(self.art, self.art)

    def _load_alias_execution_contexts(self, collection):
        execution_contexts = []

        for result in collection.values():
            alias_uuid = load_action_yaml(
                result._archiver.path)['action']['alias-of']
            execution_contexts.append(load_action_yaml(
                self.cache.data / alias_uuid)
                ['execution']['execution_context'])

        return execution_contexts



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import qiime2.core.archive.format.v5 as v5


class ArchiveFormat(v5.ArchiveFormat):
    # - Adds execution_context to the execution section of action.yaml.
    #   This looks like:
    #
    #   execution_context:
    #       type: parsl/synchronous/asynchronous
    #       parsl_type (if type is parsl): Type of executor
    #
    #   NOTE: Import actions will not have an execution_context section
    #
    # - Adds support for output collections.
    #   This looks like:
    #
    #   output-name:
    #   - The name of the entire output collection (the qiime output name)
    #   - The key of this element in the collection
    #   - The index of this element in the collection ex. '5/10' for the 5th
    #     out of 10 elements in the collection
    #
    #   Input collections now look like:
    #
    #   input-name:
    #   - key: value
    #   - key: value
    #   etc. for n elements
    pass



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import qiime2.core.transform as transform

from qiime2.core.type.util import is_visualization_type, is_collection_type


class Proxy:
    """Base class to indicate that a given class that inherits from it is a
    proxy. Also implements some generic functionality
    """
    def __eq__(self, other):
        return self.result() == other.result()

    def __ne__(self, other):
        return not (self == other)


class ProxyResult(Proxy):
    def __init__(self, future, selector, qiime_type=None):
        """We have a future that represents the results of some QIIME 2 action,
        and we have a selector indicating specifically which result we want
        """
        self._future_ = future
        self._selector_ = selector
        self._qiime_type_ = qiime_type

    def __repr__(self):
        if self._qiime_type_ is None:
            return f'<{self.__class__.__name__.lower()}: Unknown Type ' \
                   f'{object.__repr__(self)}>'
        else:
            return f'<{self.__class__.__name__.lower()}: {self.type}>'

    def __hash__(self):
        return hash(self.uuid)

    @property
    def _archiver(self):
        return self.result()._archiver

    @property
    def type(self):
        if self._qiime_type_ is not None:
            return self._qiime_type_

        return self.result().type

    @property
    def uuid(self):
        return self._archiver.uuid

    @property
    def format(self):
        from qiime2.sdk import PluginManager

        pm = PluginManager()
        return pm.get_directory_format(self.type)

    @property
    def citations(self):
        return self._archiver.citations

    def result(self):
        return self._get_element_(self._future_.result())

    def _get_element_(self, results):
        """Get the result we want off of the future we have
        """
        return getattr(results, self._selector_)

    def export_data(self, output_dir):
        return self.result().export_data(output_dir)

    def save(self, filepath, ext=None):
        """Blocks then calls save on the result.
        """
        return self.result().save(filepath, ext=ext)

    def validate(self, level=NotImplemented):
        return self.result().validate(level=level)


class ProxyArtifact(ProxyResult):
    """This represents a future Artifact that is being returned by a Parsl app
    """
    def view(self, type):
        """If we want to view the result we need the future to be resolved
        """
        return self._get_element_(self._future_.result()).view(type)

    def has_metadata(self):
        from qiime2 import Metadata

        from_type = transform.ModelType.from_view_type(self.format)
        to_type = transform.ModelType.from_view_type(Metadata)
        return from_type.has_transformation(to_type)

    def validate(self, level='max'):
        self.result().validate(level=level)


class ProxyVisualization(ProxyResult):
    """This represents a future Visualization that is being returned by a Parsl
       app
    """
    def get_index_paths(self, relative=True):
        return self.result().get_index_paths(relative=relative)


class ProxyResultCollection(Proxy):
    def __init__(self, future, selector, qiime_type=None):
        self._future_ = future
        self._selector_ = selector
        self._qiime_type_ = qiime_type

    def __len__(self):
        return len(self.collection)

    def __iter__(self):
        yield self.collection.__iter__()

    def __setitem__(self, key, item):
        self.collection[key] = item

    def __getitem__(self, key):
        return self.collection[key]

    def __repr__(self):
        return f"<{self.__class__.__name__.lower()}: {self.type}>"

    @property
    def type(self):
        # I'm not a huge fan of the fact that this may or may not need to
        # block. If this is a return from an action (which it basically
        # always will be) we don't need to block for type. Otherwise we do.
        if self._qiime_type_ is not None:
            return self._qiime_type_

        return self.result().type

    @property
    def collection(self):
        return self.result().collection

    def save(self, directory):
        """Blocks then calls save on the result.
        """
        return self.result().save(directory)

    def keys(self):
        return self.collection.keys()

    def values(self):
        return self.collection.values()

    def items(self):
        return self.collection.items()

    def _get_element_(self, results):
        """Get the result we want off of the future we have
        """
        return getattr(results, self._selector_)

    def result(self):
        return self._get_element_(self._future_.result())


class ProxyResults(Proxy):
    """This represents future results that are being returned by a Parsl app
    """
    def __init__(self, future, signature):
        """We have the future results and the outputs portion of the signature
        of the action creating the results
        """
        self._future_ = future
        self._signature_ = signature

    def __iter__(self):
        """Give us a ProxyArtifact for each result in the future
        """
        for s in self._signature_:
            yield self._create_proxy(s)

    def __getattr__(self, attr):
        """Get a particular ProxyArtifact out of the future
        """
        return self._create_proxy(attr)

    def __getitem__(self, index):
        return self._create_proxy(list(self._signature_.keys())[index])

    def __repr__(self):
        lines = []
        lines.append('%s (name = value)' % self.__class__.__name__)
        lines.append('')

        max_len = -1
        for field in self._signature_:
            if len(field) > max_len:
                max_len = len(field)

        for field, value in zip(self._signature_, self):
            field_padding = ' ' * (max_len - len(field))
            lines.append('%s%s = %r' % (field, field_padding, value))

        max_len = -1
        for line in lines:
            if len(line) > max_len:
                max_len = len(line)
        lines[1] = '-' * max_len

        return '\n'.join(lines)

    def __eq__(self, other):
        """ Overriding the one on Proxy because we have _result not result
        """
        return self._result() == other._result()

    def _asdict(self):
        return self.result()._asdict()

    def _result(self):
        """ If you are calling an action in a try-except block in a pipeline,
            you need to call this method on the Results object returned by the
            action.

            This is because if the Pipeline was executed with parsl, we need to
            block on the action in the try-except to ensure we get the result
            and raise the potential exception while we are still inside of the
            try-except. Otherwise we would just get the exception whenever the
            future resolved which would likely be outside of the try-except, so
            the exception would be raised and not caught.

            If you call an action in the Python API using parsl inside of a
            context manager (a withed in Cache for instance) you also must call
            this method there to ensure you get you don't start using a
            different cache/pool/whatever before your future resolves.
        """
        return self._future_.result()

    def _create_proxy(self, selector):
        qiime_type = self._signature_[selector].qiime_type

        if is_collection_type(qiime_type):
            return ProxyResultCollection(
                self._future_, selector, qiime_type)
        elif is_visualization_type(qiime_type):
            return ProxyVisualization(
                self._future_, selector, qiime_type)

        return ProxyArtifact(self._future_, selector, qiime_type)



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import bibtexparser as bp
import networkx as nx
import os
import pathlib
import shutil
import tempfile
import unittest
from unittest.mock import patch
import zipfile

from qiime2 import Artifact
from qiime2.sdk import PluginManager
from qiime2.sdk.usage import Usage, UsageVariable
from qiime2.plugins import ArtifactAPIUsageVariable

from ..parse import ProvDAG
from ..replay import (
    ActionCollections, BibContent, ReplayConfig, ReplayNamespaces,
    UsageVariableRecord,
    build_no_provenance_node_usage, build_import_usage, build_action_usage,
    build_usage_examples, collect_citations, dedupe_citations,
    dump_recorded_md_file, group_by_action, init_md_from_artifacts,
    init_md_from_md_file, init_md_from_recorded_md, replay_provenance,
    replay_citations
)
from .testing_utilities import CustomAssertions, DummyArtifacts
from ..usage_drivers import ReplayPythonUsage
from ...provenance import MetadataInfo

from qiime2.sdk.util import camel_to_snake


class ReplayNamespacesTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_make_unique_name(self):
        ns = ReplayNamespaces()
        self.assertEqual('name_0', ns._make_unique_name('name'))
        ns.add_usg_var_record('uuid1', 'name')
        self.assertEqual('name_1', ns._make_unique_name('name'))

    def test_get_usg_var_uuid(self):
        ns = ReplayNamespaces()
        ns.add_usg_var_record('uuid1', 'name')
        self.assertEqual('uuid1', ns.get_usg_var_uuid('name_0'))

    def test_add_usage_var_workflow(self):
        """
        Smoke tests a common workflow with this data structure
        - Create a unique variable name by adding to the namespace
        - Create a UsageVariable with that name
        - use the name to get the UUID (when we have Results, we have no UUIDs)
        - add the correctly-named UsageVariable to the namespace
        """
        use = Usage()
        uuid = self.das.concated_ints.uuid
        base_name = 'concated_ints'
        exp_name = base_name + '_0'
        ns = ReplayNamespaces()
        ns.add_usg_var_record(uuid, base_name)
        self.assertEqual(ns.get_usg_var_record(uuid).name, exp_name)

        def factory():  # pragma: no cover
            return Artifact.load(self.das.concated_ints.filepath)

        u_var = use.init_artifact(ns.get_usg_var_record(uuid).name, factory)
        self.assertEqual(u_var.name, exp_name)

        actual_uuid = ns.get_usg_var_uuid(u_var.name)
        self.assertEqual(actual_uuid, uuid)

        ns.update_usg_var_record(uuid, u_var)
        self.assertIsInstance(
            ns.get_usg_var_record(uuid).variable, UsageVariable
        )
        self.assertEqual(ns.get_usg_var_record(uuid).name, exp_name)


class ReplayProvenanceTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_replay_from_fp(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            out_fp = pathlib.Path(tmpdir) / 'rendered.txt'
            out_fn = str(out_fp)
            in_fn = self.das.concated_ints_with_md.filepath
            replay_provenance(
                ReplayPythonUsage, in_fn, out_fn, md_out_dir=tmpdir
            )

            self.assertTrue(out_fp.is_file())

            with open(out_fn, 'r') as fp:
                rendered = fp.read()

            self.assertIn('from qiime2 import Artifact', rendered)
            self.assertIn('from qiime2 import Metadata', rendered)
            self.assertIn(
                'import qiime2.plugins.dummy_plugin.actions as '
                'dummy_plugin_actions',
                rendered
            )
            self.assertIn('mapping_0 = Artifact.import_data(', rendered)
            self.assertRegex(rendered,
                             'The following command.*additional metadata')
            self.assertIn('mapping_0.view(Metadata)', rendered)
            self.assertIn('dummy_plugin_actions.identity_with_metadata',
                          rendered)
            self.assertIn('dummy_plugin_actions.concatenate_ints', rendered)

    def test_replay_from_fp_use_md_without_parse(self):
        in_fp = self.das.concated_ints.filepath
        with self.assertRaisesRegex(
                ValueError, "Metadata not parsed for replay. Re-run"
        ):
            replay_provenance(
                ReplayPythonUsage,
                in_fp,
                'unused_fp',
                parse_metadata=False,
                use_recorded_metadata=True
            )

    def test_replay_dump_md_without_parse(self):
        in_fp = self.das.concated_ints.filepath
        with self.assertRaisesRegex(
                ValueError, "(?s)Metadata not parsed,.*dump_recorded_meta"
        ):
            replay_provenance(
                ReplayPythonUsage,
                in_fp,
                'unused_fp',
                parse_metadata=False,
                dump_recorded_metadata=True
            )

    def test_replay_md_out_dir_without_parse(self):
        in_fp = self.das.concated_ints.filepath
        with self.assertRaisesRegex(
                ValueError, "(?s)Metadata not parsed,.*not.*metadata output"
        ):
            replay_provenance(
                ReplayPythonUsage,
                in_fp,
                'unused_fp',
                parse_metadata=False,
                dump_recorded_metadata=False,
                md_out_dir='/user/dumb/some_filepath'
            )

    def test_replay_use_md_without_dump_md(self):
        in_fp = self.das.concated_ints.filepath
        with self.assertRaisesRegex(
                NotImplementedError,
                "(?s)uses.*metadata.*must.*written to disk"
        ):
            replay_provenance(
                ReplayPythonUsage,
                in_fp,
                'unused_fp',
                use_recorded_metadata=True,
                dump_recorded_metadata=False
            )

    def test_replay_from_provdag(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            out_fp = pathlib.Path(tmpdir) / 'rendered.txt'
            out_fn = str(out_fp)
            dag = self.das.concated_ints_with_md.dag
            replay_provenance(ReplayPythonUsage, dag, out_fn,
                              md_out_dir=tmpdir)

            self.assertTrue(out_fp.is_file())

            with open(out_fn, 'r') as fp:
                rendered = fp.read()

            self.assertIn('from qiime2 import Artifact', rendered)
            self.assertIn('from qiime2 import Metadata', rendered)
            self.assertIn(
                'import qiime2.plugins.dummy_plugin.actions as '
                'dummy_plugin_actions',
                rendered
            )
            self.assertIn('mapping_0 = Artifact.import_data(', rendered)
            self.assertRegex(rendered,
                             'The following command.*additional metadata')
            self.assertIn('mapping_0.view(Metadata)', rendered)
            self.assertIn('dummy_plugin_actions.identity_with_metadata',
                          rendered)
            self.assertIn('dummy_plugin_actions.concatenate_ints', rendered)

    def test_replay_from_provdag_use_md_without_parse(self):
        dag = ProvDAG(self.das.concated_ints_with_md.filepath,
                      validate_checksums=False,
                      parse_metadata=False)
        with self.assertRaisesRegex(
                ValueError, "Metadata not parsed for replay"
        ):
            replay_provenance(
                ReplayPythonUsage, dag, 'unused', use_recorded_metadata=True
            )

    def test_replay_from_provdag_ns_collision(self):
        """
        This artifact's dag contains a few results with the output-name
        filtered-table, so is a good check for namespace collisions if
        we're not uniquifying variable names properly.
        """
        with tempfile.TemporaryDirectory() as tempdir:
            self.das.concated_ints.artifact.save(
                os.path.join(tempdir, 'c1.qza'))
            self.das.other_concated_ints.artifact.save(
                os.path.join(tempdir, 'c2.qza'))
            dag = ProvDAG(tempdir)

        exp = ['concatenated_ints_0', 'concatenated_ints_1']
        with tempfile.TemporaryDirectory() as tempdir:
            out_path = pathlib.Path(tempdir) / 'ns_coll.txt'
            replay_provenance(
                 ReplayPythonUsage, dag, out_path, md_out_dir=tempdir
            )

            with open(out_path, 'r') as fp:
                rendered = fp.read()
                for name in exp:
                    self.assertIn(name, rendered)

    def test_replay_optional_param_is_none(self):
        dag = self.das.int_seq_optional_input.dag
        with tempfile.TemporaryDirectory() as tempdir:
            out_path = pathlib.Path(tempdir) / 'ns_coll.txt'

            replay_provenance(
                ReplayPythonUsage, dag, out_path, md_out_dir=tempdir
            )
            with open(out_path, 'r') as fp:
                rendered = fp.read()
            self.assertIn('ints=int_sequence1_0', rendered)
            self.assertIn('num1=', rendered)
            self.assertNotIn('optional1=', rendered)
            self.assertNotIn('num2=', rendered)


class MultiplePluginTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        from qiime2.sdk.plugin_manager import PluginManager
        from qiime2 import Artifact

        cls.pm = PluginManager()
        cls.dp = cls.pm.plugins['dummy-plugin']
        cls.op = cls.pm.plugins['other-plugin']
        cls.tempdir = tempfile.mkdtemp(prefix='qiime2-other-plugin-temp-')

        int_seq = Artifact.import_data('IntSequence1', [1, 2, 3, 4])
        concat_ints = cls.op.methods['concatenate_ints']
        split_ints = cls.dp.methods['split_ints']

        concated_ints, = concat_ints(
            int_seq, int_seq, int_seq, 5, 6
        )
        cls.splitted_ints, _ = split_ints(concated_ints)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tempdir)

    def test_multiple_plugins_in_provenance(self):
        fp = os.path.join(self.tempdir, 'splitted_ints.qza')
        self.splitted_ints.save(fp)

        with tempfile.TemporaryDirectory() as tempdir:
            out_fp = os.path.join(tempdir, 'rendered.txt')
            replay_provenance(
                ReplayPythonUsage, fp, out_fp, md_out_dir=tempdir
            )

            with open(out_fp, 'r') as fp:
                rendered = fp.read()

        self.assertIn('from qiime2 import Artifact', rendered)
        self.assertIn(
            'import qiime2.plugins.dummy_plugin.actions as '
            'dummy_plugin_actions',
            rendered
        )
        self.assertIn(
            'import qiime2.plugins.other_plugin.actions as '
            'other_plugin_actions',
            rendered
        )
        self.assertIn(
            'dummy_plugin_actions.split_ints(', rendered
        )
        self.assertIn(
            'other_plugin_actions.concatenate_ints(', rendered
        )


class ReplayProvDAGDirectoryTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_directory_replay_multiple_imports(self):
        """
        The directory being parsed here contains two pairs of duplicates,
        and should replay as only two import statements.
        """
        outer_dir = os.path.join(self.tempdir, 'outer')
        inner_dir = os.path.join(outer_dir, 'inner')
        os.makedirs(inner_dir)

        for artifact in self.das.single_int, self.das.single_int2:
            for dir_ in inner_dir, outer_dir:
                artifact.artifact.save(
                    os.path.join(dir_, f'{artifact.name}.qza')
                )

        dir_dag = ProvDAG(outer_dir)
        self.assertEqual(len(dir_dag._parsed_artifact_uuids), 2)
        self.assertIn(self.das.single_int.uuid, dir_dag.dag)
        self.assertIn(self.das.single_int2.uuid, dir_dag.dag)

        exp_1 = (
            '(?s)from qiime2 import Artifact.*'
            'single_int_0 = Artifact.import_data.*'
            '<your data here>.*'
        )
        exp_2 = (
            '(?s)from qiime2 import Artifact.*'
            'single_int_1 = Artifact.import_data.*'
            '<your data here>.*'
        )

        with tempfile.TemporaryDirectory() as tempdir:
            out_path = pathlib.Path(tempdir) / 'rendered.txt'
            replay_provenance(
                ReplayPythonUsage, dir_dag, out_path, md_out_dir=tempdir
            )
            self.assertTrue(out_path.is_file())

            with open(out_path, 'r') as fp:
                rendered = fp.read()
                self.assertRegex(rendered, exp_1)
                self.assertRegex(rendered, exp_2)


class BuildUsageExamplesTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir
        cls.pm = PluginManager()

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    @patch('qiime2.core.archive.provenance_lib.replay.build_action_usage')
    @patch('qiime2.core.archive.provenance_lib.replay.build_import_usage')
    @patch('qiime2.core.archive.provenance_lib.replay.'
           'build_no_provenance_node_usage')
    def test_build_usage_examples(self, n_p_builder, imp_builder, act_builder):
        ns = ReplayNamespaces()
        dag = self.das.concated_ints_with_md.dag
        cfg = ReplayConfig(
            use=ReplayPythonUsage(),
            use_recorded_metadata=False,
            pm=self.pm
        )
        build_usage_examples(dag, cfg, ns)

        n_p_builder.assert_not_called()
        self.assertEqual(imp_builder.call_count, 3)
        self.assertEqual(act_builder.call_count, 2)

    @patch('qiime2.core.archive.provenance_lib.replay.build_action_usage')
    @patch('qiime2.core.archive.provenance_lib.replay.build_import_usage')
    @patch('qiime2.core.archive.provenance_lib.replay.'
           'build_no_provenance_node_usage')
    def test_build_usage_examples_lone_v0(
            self, n_p_builder, imp_builder, act_builder
    ):
        ns = ReplayNamespaces()
        uuid = self.das.table_v0.uuid
        with self.assertWarnsRegex(
                UserWarning, f'(:?)Art.*{uuid}.*prior.*incomplete'
        ):
            dag = ProvDAG(self.das.table_v0.filepath)

        cfg = ReplayConfig(
            use=ReplayPythonUsage(),
            use_recorded_metadata=False,
            pm=self.pm
        )
        build_usage_examples(dag, cfg, ns)

        # This is a single v0 archive, so should have only one np node
        n_p_builder.assert_called_once()
        imp_builder.assert_not_called()
        act_builder.assert_not_called()

    @patch('qiime2.core.archive.provenance_lib.replay.build_action_usage')
    @patch('qiime2.core.archive.provenance_lib.replay.build_import_usage')
    @patch('qiime2.core.archive.provenance_lib.replay.'
           'build_no_provenance_node_usage')
    def test_build_usage_examples_mixed(
            self, n_p_builder, imp_builder, act_builder
    ):
        mixed_dir = os.path.join(self.tempdir, 'mixed-dir')
        os.mkdir(mixed_dir)
        shutil.copy(self.das.table_v0.filepath, mixed_dir)
        shutil.copy(self.das.concated_ints_v6.filepath, mixed_dir)

        ns = ReplayNamespaces()
        v0_uuid = self.das.table_v0.uuid
        with self.assertWarnsRegex(
                UserWarning, f'(:?)Art.*{v0_uuid}.*prior.*incomplete'
        ):
            dag = ProvDAG(mixed_dir)

        cfg = ReplayConfig(
            use=ReplayPythonUsage(),
            use_recorded_metadata=False,
            pm=self.pm
        )
        build_usage_examples(dag, cfg, ns)

        n_p_builder.assert_called_once()
        self.assertEqual(imp_builder.call_count, 2)
        act_builder.assert_called_once()

    @patch('qiime2.core.archive.provenance_lib.replay.build_action_usage')
    @patch('qiime2.core.archive.provenance_lib.replay.build_import_usage')
    @patch('qiime2.core.archive.provenance_lib.replay.'
           'build_no_provenance_node_usage')
    def test_build_usage_examples_big(
            self, n_p_builder, imp_builder, act_builder):

        many_dir = os.path.join(self.tempdir, 'many-dir')
        os.mkdir(many_dir)
        shutil.copy(self.das.concated_ints_with_md.filepath, many_dir)
        shutil.copy(self.das.splitted_ints.filepath, many_dir)
        shutil.copy(self.das.pipeline_viz.filepath, many_dir)

        ns = ReplayNamespaces()
        dag = ProvDAG(many_dir)
        cfg = ReplayConfig(
            use=ReplayPythonUsage(),
            use_recorded_metadata=False,
            pm=self.pm
        )
        build_usage_examples(dag, cfg, ns)

        n_p_builder.assert_not_called()
        # concated_ints_with_md is loaded from disk so imports don't overlap
        # with splitted_ints and pipeline_viz
        self.assertEqual(imp_builder.call_count, 6)
        self.assertEqual(act_builder.call_count, 4)


class MiscHelperFnTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir
        cls.pm = PluginManager()

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_uniquify_action_name(self):
        ns = ReplayNamespaces()
        p1 = 'dummy_plugin'
        a1 = 'action_jackson'
        p2 = 'dummy_plugin'
        a2 = 'missing_in_action'

        unique1 = ns.uniquify_action_name(p1, a1)
        self.assertEqual(unique1, 'dummy_plugin_action_jackson_0')

        unique2 = ns.uniquify_action_name(p2, a2)
        self.assertEqual(unique2, 'dummy_plugin_missing_in_action_0')

        duplicate = ns.uniquify_action_name(p1, a1)
        self.assertEqual(duplicate, 'dummy_plugin_action_jackson_1')

    def test_dump_recorded_md_file_no_md(self):
        uuid = self.das.table_v0.uuid
        dag = self.das.table_v0.dag

        cfg = ReplayConfig(use=ReplayPythonUsage(),
                           pm=self.pm)
        provnode = dag.get_node_data(uuid)
        action_name = 'old_action'
        md_id = 'metadata'
        fn = 'metadata.tsv'

        with self.assertRaisesRegex(
            ValueError, "should only be called.*if.*metadata"
        ):
            dump_recorded_md_file(cfg, provnode, action_name, md_id, fn)


class GroupByActionTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir
        cls.pm = PluginManager()

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_gba_with_provenance(self):
        self.maxDiff = None

        ns = ReplayNamespaces()
        dag = self.das.concated_ints_v6.dag
        sorted_nodes = nx.topological_sort(dag.collapsed_view)
        actual = group_by_action(dag, sorted_nodes, ns)
        exp = {
            'b49e497c-19b2-49f7-b9a2-0d837016c151': {
                '8dea2f1a-2164-4a85-9f7d-e0641b1db22b': 'int_sequence1'
            },
            '12988290-1ebf-47ad-8c34-5469d42e5ffe': {
                '7727c060-5384-445d-b007-b64b41a090ee': 'int_sequence2'
            },
            '5035a60e-6f9a-40d4-b412-48ae52255bb5': {
                '6facaf61-1676-45eb-ada0-d530be678b27': 'concatenated_ints'
            }
        }
        self.assertEqual(actual.std_actions, exp)
        self.assertEqual(actual.no_provenance_nodes, [])

    def test_gba_no_provenance(self):
        ns = ReplayNamespaces()
        dag = self.das.table_v0.dag
        uuid = self.das.table_v0.uuid

        sorted_nodes = nx.topological_sort(dag.collapsed_view)
        action_collections = group_by_action(dag, sorted_nodes, ns)
        self.assertEqual(action_collections.std_actions, {})
        self.assertEqual(action_collections.no_provenance_nodes, [uuid])

    def test_gba_some_nodes_missing_provenance(self):
        mixed_dir = os.path.join(self.tempdir, 'mixed-dir')
        os.mkdir(mixed_dir)
        shutil.copy(self.das.table_v0.filepath, mixed_dir)
        shutil.copy(self.das.concated_ints_v6.filepath, mixed_dir)

        ns = ReplayNamespaces()
        v0_uuid = self.das.table_v0.uuid
        with self.assertWarnsRegex(
                UserWarning, f'(:?)Art.*{v0_uuid}.*prior.*incomplete'
        ):
            dag = ProvDAG(mixed_dir)

        sorted_nodes = nx.topological_sort(dag.collapsed_view)
        action_collections = group_by_action(dag, sorted_nodes, ns)

        exp = {
            'b49e497c-19b2-49f7-b9a2-0d837016c151': {
                '8dea2f1a-2164-4a85-9f7d-e0641b1db22b': 'int_sequence1'
            },
            '12988290-1ebf-47ad-8c34-5469d42e5ffe': {
                '7727c060-5384-445d-b007-b64b41a090ee': 'int_sequence2'
            },
            '5035a60e-6f9a-40d4-b412-48ae52255bb5': {
                '6facaf61-1676-45eb-ada0-d530be678b27': 'concatenated_ints'
            }
        }
        self.assertEqual(action_collections.std_actions, exp)
        self.assertEqual(action_collections.no_provenance_nodes, [v0_uuid])


class InitializerTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir
        cls.pm = PluginManager()

        with zipfile.ZipFile(cls.das.concated_ints_with_md.filepath) as zf:
            root_node_id = cls.das.concated_ints_with_md.uuid
            all_filenames = zf.namelist()
            dag = cls.das.concated_ints_with_md.dag
            for node in dag.nodes:
                md_path = os.path.join(
                    root_node_id, 'provenance', 'artifacts', node, 'action',
                    'metadata.tsv'
                )
                if md_path in all_filenames:
                    cls.md_node_id = node
                else:
                    cls.non_md_node_id = node

        with zipfile.ZipFile(
            cls.das.concated_ints_with_md_column.filepath
        ) as zf:
            root_node_id = cls.das.concated_ints_with_md_column.uuid
            all_filenames = zf.namelist()
            dag = cls.das.concated_ints_with_md_column.dag
            for node in dag.nodes:
                md_path = os.path.join(
                    root_node_id, 'provenance', 'artifacts', node, 'action',
                    'metadata.tsv'
                )
                if md_path in all_filenames:
                    cls.mdc_node_id = node
                else:
                    cls.non_mdc_node_id = node

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_init_md_from_artifacts_no_artifacts(self):
        cfg = ReplayConfig(
            use=ReplayPythonUsage(),
            use_recorded_metadata=False,
            pm=self.pm
        )
        ns = ReplayNamespaces

        # create dummy hash '0', not relevant here
        md_info = MetadataInfo([], 'hmm.tsv', '0')
        with self.assertRaisesRegex(
            ValueError, "not.*used.*input_artifact_uuids.*empty"
        ):
            init_md_from_artifacts(md_info, ns, cfg)

    def test_init_md_from_artifacts_one_art(self):
        # This helper doesn't capture real data, so we're only smoke testing,
        # checking type, and confirming the repr looks reasonable.
        cfg = ReplayConfig(
            use=ReplayPythonUsage(),
            use_recorded_metadata=False,
            pm=self.pm
        )

        # We expect artifact vars have already been added to the namespace
        a1 = cfg.use.init_artifact(name='thing1', factory=lambda: None)
        ns = ReplayNamespaces()
        ns._usg_var_ns = {'uuid1': UsageVariableRecord('thing1', a1)}

        # create dummy hash '0', not relevant here
        md_info = MetadataInfo(['uuid1'], 'hmm.tsv', '0')
        var = init_md_from_artifacts(md_info, ns, cfg)
        self.assertIsInstance(var, UsageVariable)
        self.assertEqual(var.var_type, 'metadata')
        rendered = var.use.render()

        self.assertIn('from qiime2 import Metadata', rendered)
        self.assertIn('thing1_a_0_md = thing1.view(Metadata)', rendered)

    def test_init_md_from_artifacts_many(self):
        # This helper doesn't capture real data, so we're only smoke testing,
        # checking type, and confirming the repr looks reasonable.
        cfg = ReplayConfig(
            use=ReplayPythonUsage(),
            use_recorded_metadata=False,
            pm=self.pm
        )

        # We expect artifact vars have already been added to the namespace
        a1 = cfg.use.init_artifact(name='thing1', factory=lambda: None)
        a2 = cfg.use.init_artifact(name='thing2', factory=lambda: None)
        a3 = cfg.use.init_artifact(name='thing3', factory=lambda: None)

        ns = ReplayNamespaces()
        ns._usg_var_ns = {
            'uuid1': UsageVariableRecord('thing1', a1),
            'uuid2': UsageVariableRecord('thing2', a2),
            'uuid3': UsageVariableRecord('thing3', a3),
        }

        # create dummy hash '0', not relevant here
        md_info = MetadataInfo(['uuid1', 'uuid2', 'uuid3'], 'hmm.tsv', '0')
        var = init_md_from_artifacts(md_info, ns, cfg)
        self.assertIsInstance(var, UsageVariable)
        self.assertEqual(var.var_type, 'metadata')
        rendered = var.use.render()

        self.assertIn('from qiime2 import Metadata', rendered)
        self.assertIn('thing1_a_0_md = thing1.view(Metadata)', rendered)
        self.assertIn('thing2_a_0_md = thing2.view(Metadata)', rendered)
        self.assertIn('thing3_a_0_md = thing3.view(Metadata)', rendered)
        self.assertIn(
            'merged_artifacts_0_md = '
            'thing1_a_0_md.merge(thing2_a_0_md, thing3_a_0_md)',
            rendered
        )

    def test_init_md_from_md_file(self):
        dag = self.das.concated_ints_with_md.dag
        md_node = dag.get_node_data(self.md_node_id)
        md_id = 'whatevs'
        param_name = 'metadata'

        ns = ReplayNamespaces()
        ns.add_usg_var_record(md_id, param_name)

        cfg = ReplayConfig(
            use=ReplayPythonUsage(),
            use_recorded_metadata=False,
            pm=self.pm
        )

        var = init_md_from_md_file(md_node, param_name, md_id, ns, cfg)

        rendered = var.use.render()
        self.assertIn('from qiime2 import Metadata', rendered)
        self.assertIn(
            'metadata_0_md = Metadata.load(<your metadata filepath>)',
            rendered
        )

    def test_init_md_from_recorded_md(self):
        dag = self.das.concated_ints_with_md.dag
        no_md_node = dag.get_node_data(self.non_md_node_id)
        md_node = dag.get_node_data(self.md_node_id)
        var_name = 'metadata_0'
        param_name = 'metadata'

        ns = ReplayNamespaces()
        ns.add_usg_var_record(var_name, param_name)

        cfg = ReplayConfig(
            use=ReplayPythonUsage(),
            use_recorded_metadata=False,
            pm=self.pm
        )

        md_fn = 'identity_with_metadata/metadata_0'

        with self.assertRaisesRegex(ValueError, 'only.*call.*if.*metadata'):
            init_md_from_recorded_md(
                no_md_node, param_name, var_name, ns, cfg, md_fn
            )

        var = init_md_from_recorded_md(
            md_node, param_name, var_name, ns, cfg, md_fn
        )
        self.assertIsInstance(var, UsageVariable)
        self.assertEqual(var.var_type, 'metadata')

        rendered = cfg.use.render()
        self.assertIn('from qiime2 import Metadata', rendered)
        self.assertIn('metadata_0_md = Metadata.load', rendered)
        self.assertIn(
            'recorded_metadata/identity_with_metadata/metadata_0', rendered
        )

    def test_init_md_from_recorded_mdc(self):
        dag = self.das.concated_ints_with_md_column.dag
        no_md_node = dag.get_node_data(self.non_mdc_node_id)
        md_node = dag.get_node_data(self.mdc_node_id)
        var_name = 'metadata_0'
        param_name = 'metadata'

        ns = ReplayNamespaces()
        ns.add_usg_var_record(var_name, param_name)

        cfg = ReplayConfig(
            use=ReplayPythonUsage(),
            use_recorded_metadata=False,
            pm=self.pm
        )

        md_fn = 'identity_with_metadata_column/metadata_0'

        with self.assertRaisesRegex(ValueError, 'only.*call.*if.*metadata'):
            init_md_from_recorded_md(
                no_md_node, param_name, var_name, ns, cfg, md_fn
            )

        var = init_md_from_recorded_md(
            md_node, param_name, var_name, ns, cfg, md_fn
        )
        self.assertIsInstance(var, UsageVariable)
        self.assertEqual(var.var_type, 'column')

        rendered = cfg.use.render()
        self.assertIn('from qiime2 import Metadata', rendered)
        self.assertIn('metadata_0_md = Metadata.load', rendered)
        self.assertIn('.get_column(', rendered)
        self.assertIn('recorded_metadata/identity_with_metadata_column/'
                      'metadata_0.tsv', rendered)


class BuildNoProvenanceUsageTests(CustomAssertions):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir
        cls.pm = PluginManager()

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_build_no_provenance_node_usage_w_complete_node(self):
        ns = ReplayNamespaces()
        cfg = ReplayConfig(use=ReplayPythonUsage(),
                           use_recorded_metadata=False, pm=self.pm)
        uuid = self.das.table_v0.uuid
        dag = self.das.table_v0.dag
        v0_node = dag.get_node_data(uuid)
        build_no_provenance_node_usage(v0_node, uuid, ns, cfg)

        out_var_name = 'feature_table_frequency_0'
        self.assertIn(uuid, ns._usg_var_ns)
        self.assertEqual(ns._usg_var_ns[uuid].name, out_var_name)

        rendered = cfg.use.render()
        # Confirm the initial context comment is present once.
        self.assertREAppearsOnlyOnce(rendered, 'nodes have no provenance')
        header = '# Original Node ID                       String Description'
        self.assertREAppearsOnlyOnce(rendered, header)

        # Confirm expected values have been rendered
        exp_v0 = f'# {uuid}   feature_table_frequency_0'
        self.assertRegex(rendered, exp_v0)

    def test_build_no_provenance_node_usage_uuid_only_node(self):
        ns = ReplayNamespaces()
        cfg = ReplayConfig(
            use=ReplayPythonUsage(),
            use_recorded_metadata=False,
            pm=self.pm
        )

        uuid = 'some-uuid'
        node = None
        build_no_provenance_node_usage(node, uuid, ns, cfg)

        out_var_name = 'no-provenance-node_0'
        self.assertIn(uuid, ns._usg_var_ns)
        self.assertEqual(ns._usg_var_ns[uuid].name, out_var_name)

        rendered = cfg.use.render()
        # Confirm the initial context comment is present once.
        self.assertREAppearsOnlyOnce(rendered, 'nodes have no provenance')
        header = '# Original Node ID                       String Description'
        self.assertREAppearsOnlyOnce(rendered, header)

        # Confirm expected values have been rendered
        exp_v0 = f'# {uuid}   no_provenance_node_0'
        self.assertRegex(rendered, exp_v0)

    def test_build_no_provenance_node_usage_many(self):
        ns = ReplayNamespaces()
        cfg = ReplayConfig(
            use=ReplayPythonUsage(),
            use_recorded_metadata=False, pm=self.pm
        )

        # This function doesn't actually know about the DAG, so no need to join
        uuid = self.das.table_v0.uuid
        dag = self.das.table_v0.dag
        v0_node = dag.get_node_data(uuid)

        dummy_node_uuid = uuid + '-dummy'
        dummy_node = dag.get_node_data(uuid)

        build_no_provenance_node_usage(v0_node, uuid, ns, cfg)
        build_no_provenance_node_usage(dummy_node, dummy_node_uuid, ns, cfg)
        self.assertIn(uuid, ns._usg_var_ns)
        self.assertIn(dummy_node_uuid, ns._usg_var_ns)
        self.assertEqual(
            ns._usg_var_ns[uuid].name, 'feature_table_frequency_0'
        )
        self.assertEqual(
            ns._usg_var_ns[dummy_node_uuid].name, 'feature_table_frequency_1'
        )

        rendered = cfg.use.render()
        # Confirm the initial context isn't repeated.
        self.assertREAppearsOnlyOnce(rendered, 'nodes have no provenance')
        header = '# Original Node ID                       String Description'
        self.assertREAppearsOnlyOnce(rendered, header)

        # Confirm expected values have been rendered
        exp_og = f'# {uuid}   feature_table_frequency_0'
        exp_dummy = f'# {uuid}-dummy   feature_table_frequency_1'
        self.assertRegex(rendered, exp_og)
        self.assertRegex(rendered, exp_dummy)


class BuildImportUsageTests(CustomAssertions):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir
        cls.pm = PluginManager()

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_build_import_usage_python(self):
        ns = ReplayNamespaces()
        cfg = ReplayConfig(
            use=ReplayPythonUsage(),
            use_recorded_metadata=False,
            pm=self.pm
        )

        dag = self.das.concated_ints_v6.dag
        import_uuid = '8dea2f1a-2164-4a85-9f7d-e0641b1db22b'
        import_node = dag.get_node_data(import_uuid)
        c_to_s_type = camel_to_snake(import_node.type)
        unq_var_nm = c_to_s_type + '_0'
        build_import_usage(import_node, ns, cfg)

        usg_var = ns.get_usg_var_record(import_uuid).variable
        self.assertIsInstance(usg_var, UsageVariable)
        self.assertEqual(usg_var.var_type, 'artifact')
        self.assertEqual(usg_var.name, unq_var_nm)

        rendered = cfg.use.render()
        out_name = usg_var.to_interface_name()
        self.assertRegex(rendered, 'from qiime2 import Artifact')
        self.assertRegex(rendered, rf'{out_name} = Artifact.import_data\(')
        self.assertRegex(rendered, import_node.type)
        self.assertRegex(rendered, '<your data here>')


class ReplayResultCollectionTests(CustomAssertions):
    '''
    One of three structures may be encoutered when parsing the inputs section
    of action.yaml, described below:

    case 1 (single artifact case):

        inputs:
        - some_input_name: some_uuid
        - some_other_input_name: some_other_uuid
        (...)

        For the single artifact case we may find the artifact in the usage
        variable namespace or it might only exist in a result collection.
        - case 1a:
            The artifact exists in the usage variable namespace so access it
            from there, not from any result collection that may contain it.
        - case 1b:
            The artifact does not exist in the usage variable namespace so
            find it in a result collection, destructure it, and then use the
            destructured artifact.

    case 2 (list of artifacts case):

        inputs:
        - some_input_name:
            - some_uuid
            - some_other_uuid
        (...)

        For the list of artifacts case the list contents may be equivalent to
        an existing result collection, or it may not be.
        - case 2a:
            The list contents are not equivalent to any existing result
            collection so for each member do either case 1a or 1b as described
            above.
        - case 2b:
            The list contents are equivalent to an existing result collection
            so pass in the result collection directly (which will be cast to
            a list).

    case 3 (result collection case):

        inputs:
        - result_collection_name:
            - some_key: some_uuid
            - some_other_key: some_other_uuid
        (...)

        For the result collection case an equivalent existing result collection
        may exist or it may not.
        - case 3a:
            No equivalent result collection is found so for each member follow
            case 1a or 1b as described above.
        - case 3b:
            An equivalent result collection is found so pass it in directly.

    '''
    @classmethod
    def setUpClass(cls):
        cls.pm = PluginManager()
        cls.dp = cls.pm.plugins['dummy-plugin']

        cls.single_int = Artifact.import_data('SingleInt', 0)
        cls.dict_of_ints = cls.dp.methods['dict_of_ints']
        cls.list_of_ints = cls.dp.methods['list_of_ints']

    def test_cases_1a_1b(self):
        '''
        The `single_int` usage variable is not found before the first call to
        `optional_artifact_pipeline`, so it is accessed from the result
        collection it belongs to (case 1b) and added to the usage variable
        namespace. When we call `optional_artifact_pipeline` again it should
        not destructure the result collection a second time but instead just
        use the available usage variable that we added to the namespace
        previously (case 1a).
        '''
        int_seq = Artifact.import_data('IntSequence1', [1, 1, 2])

        opt_art_pipeline = self.dp.pipelines['optional_artifact_pipeline']

        rc = {'int1': self.single_int}

        rc_out, = self.dict_of_ints(rc)
        int_destructured = rc_out['int1']
        int_seq_out, = opt_art_pipeline(int_seq, int_destructured)
        int_seq_out, = opt_art_pipeline(int_seq_out, int_destructured)

        with tempfile.TemporaryDirectory() as tempdir:
            in_fp = pathlib.Path(tempdir) / 'int_seq_out.qza'
            int_seq_out.save(in_fp)
            dag = ProvDAG(in_fp)

            out_fp = pathlib.Path(tempdir) / 'rendered.txt'
            out_fn = str(out_fp)
            replay_provenance(ReplayPythonUsage, dag, out_fn)

            with open(out_fp) as fh:
                rendered = fh.read()

        exp = '''\
single_int_1 = output_0_artifact_collection['int1']
ints_1, = dummy_plugin_actions.optional_artifact_pipeline(
    int_sequence=int_sequence1_0,
    single_int=single_int_1,
)
# SAVE: comment out the following with '# ' to skip saving Results to disk
ints_1.save('ints_1')

ints_2, = dummy_plugin_actions.optional_artifact_pipeline(
    int_sequence=ints_1,
    single_int=single_int_1,
)
'''
        self.assertIn(exp, rendered)

    def test_case_2a(self):
        rc = {'int1': self.single_int, 'int2': self.single_int}
        list_rc_out, = self.list_of_ints(rc)
        int1_destructured = list_rc_out['0']

        with tempfile.TemporaryDirectory() as tempdir:
            in_fp = pathlib.Path(tempdir) / 'int1_destructured.qza'
            int1_destructured.save(in_fp)
            dag = ProvDAG(in_fp)

            out_fp = pathlib.Path(tempdir) / 'rendered.txt'
            out_fn = str(out_fp)
            replay_provenance(ReplayPythonUsage, dag, out_fn)

            with open(out_fp) as fh:
                rendered = fh.read()

        exp = '''\
output_0_artifact_collection, = dummy_plugin_actions.list_of_ints(
    ints=[single_int_0, single_int_0],
)
'''
        self.assertIn(exp, rendered)

    def test_case_2b(self):
        rc = {'int1': self.single_int}
        rc_out, = self.dict_of_ints(rc)
        list_rc_out, = self.list_of_ints(rc_out)
        int1_destructured = list_rc_out['0']

        with tempfile.TemporaryDirectory() as tempdir:
            in_fp = pathlib.Path(tempdir) / 'int1_destructured.qza'
            int1_destructured.save(in_fp)
            dag = ProvDAG(in_fp)

            out_fp = pathlib.Path(tempdir) / 'rendered.txt'
            out_fn = str(out_fp)
            replay_provenance(ReplayPythonUsage, dag, out_fn)

            with open(out_fp) as fh:
                rendered = fh.read()

        exp = '''\
output_1_artifact_collection, = dummy_plugin_actions.list_of_ints(
    ints=output_0_artifact_collection,
)
'''
        self.assertIn(exp, rendered)

    def test_case_3a(self):
        '''
        The `int1_destructured` artifact won't be found in the usage variable
        namespace, while `self.single_int` will be, so assert that int1 is
        destructured and `self.single_int` is just used as is because its
        variable name is already available.
        '''
        rc1 = {'int1': self.single_int}
        rc1_out, = self.dict_of_ints(rc1)
        int1_destructured = rc1_out['int1']

        rc2 = {'int1': int1_destructured, 'int2': self.single_int}
        rc2_out, = self.dict_of_ints(rc2)
        int3_destructured = rc2_out['int2']

        with tempfile.TemporaryDirectory() as tempdir:
            in_fp = pathlib.Path(tempdir) / 'int2_destructured.qza'
            int3_destructured.save(in_fp)
            dag = ProvDAG(in_fp)

            out_fp = pathlib.Path(tempdir) / 'rendered.txt'
            out_fn = str(out_fp)
            replay_provenance(ReplayPythonUsage, dag, out_fn)

            with open(out_fp) as fh:
                rendered = fh.read()

        exp = '''\
ints_1 = output_0_artifact_collection['int1']
ints_2_artifact_collection = ResultCollection({
    'int1': ints_1,
    'int2': single_int_0,
})
output_1_artifact_collection, = dummy_plugin_actions.dict_of_ints(
    ints=ints_2_artifact_collection,
)
'''
        self.assertIn(exp, rendered)
        self.assertREAppearsOnlyOnce(rendered, r'\[.*\]')

    def test_case_3b(self):
        rc = {'int1': self.single_int, 'int2': self.single_int}

        rc_out, = self.dict_of_ints(rc)
        rc_out_2, = self.dict_of_ints(rc_out)
        int1_destructured = rc_out_2['int1']

        with tempfile.TemporaryDirectory() as tempdir:
            in_fp = pathlib.Path(tempdir) / 'int1_destructured.qza'
            int1_destructured.save(in_fp)
            dag = ProvDAG(in_fp)

            out_fp = pathlib.Path(tempdir) / 'rendered.txt'
            out_fn = str(out_fp)
            replay_provenance(ReplayPythonUsage, dag, out_fn)

            with open(out_fp) as fh:
                rendered = fh.read()

        exp = '''\
output_1_artifact_collection, = dummy_plugin_actions.dict_of_ints(
    ints=output_0_artifact_collection,
)
'''
        self.assertIn(exp, rendered)
        self.assertREAppearsOnlyOnce(rendered, r'ResultCollection\(')


class BuildActionUsageTests(CustomAssertions):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir
        cls.pm = PluginManager()

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_build_action_usage_python(self):
        plugin = 'dummy_plugin'
        action = 'concatenate_ints'
        cfg = ReplayConfig(
            use=ReplayPythonUsage(),
            use_recorded_metadata=False,
            pm=self.pm
        )

        ns = ReplayNamespaces()
        import_var_1 = ArtifactAPIUsageVariable(
            'imported_ints_0', lambda: None, 'artifact', cfg.use
        )
        import_var_2 = ArtifactAPIUsageVariable(
            'imported_ints_1', lambda: None, 'artifact', cfg.use
        )
        import_uuid_1 = '8dea2f1a-2164-4a85-9f7d-e0641b1db22b'
        import_uuid_2 = '7727c060-5384-445d-b007-b64b41a090ee'
        ns.add_usg_var_record(import_uuid_1, 'imported_ints', import_var_1)
        ns.add_usg_var_record(import_uuid_2, 'imported_ints', import_var_2)

        dag = self.das.concated_ints_v6.dag
        action_uuid = '5035a60e-6f9a-40d4-b412-48ae52255bb5'
        node_uuid = '6facaf61-1676-45eb-ada0-d530be678b27'
        node = dag.get_node_data(node_uuid)
        actions = ActionCollections(
            std_actions={action_uuid: {node_uuid: 'concatenated_ints'}}
        )
        unique_var_name = node.action.output_name + '_0'
        build_action_usage(node, ns, actions.std_actions, action_uuid, cfg)

        usg_var = ns.get_usg_var_record(node_uuid).variable
        out_name = usg_var.to_interface_name()

        self.assertIsInstance(usg_var, UsageVariable)
        self.assertEqual(usg_var.var_type, 'artifact')
        self.assertEqual(usg_var.name, unique_var_name)

        rendered = cfg.use.render()
        self.assertRegex(
            rendered, f"import.*{plugin}.actions as {plugin}_actions"
        )
        self.assertIn(
            f'{out_name}, = dummy_plugin_actions.{action}(', rendered
        )

    def test_build_action_usage_recorded_md(self):
        action = 'identity_with_metadata'
        with tempfile.TemporaryDirectory() as tempdir:
            cfg = ReplayConfig(
                use=ReplayPythonUsage(),
                use_recorded_metadata=False,
                pm=self.pm,
                md_out_dir=tempdir
            )

            action_uuid = '8dae7a81-83ce-48db-9313-6e3131b0933c'
            node_uuid = 'be472b56-d205-43ee-8180-474da575c4d5'

            dag = self.das.concated_ints_with_md.dag
            node = dag.get_node_data(node_uuid)

            ns = ReplayNamespaces()
            mapping_var = ArtifactAPIUsageVariable(
                'imported_mapping_0', lambda: None, 'artifact', cfg.use
            )
            intseq_var_1 = ArtifactAPIUsageVariable(
                'imported_ints_0', lambda: None, 'artifact', cfg.use
            )
            intseq_var_2 = ArtifactAPIUsageVariable(
                'imported_ints_1', lambda: None, 'artifact', cfg.use
            )
            mapping_import_uuid = '8f71b73d-b028-4cbc-9894-738bdfe718bf'
            intseq_import_uuid_1 = '0bb6d731-155a-4dd0-8a1e-98827bc4e0bf'
            intseq_import_uuid_2 = 'e6b37bae-3a14-40f7-87b4-52cf5c7c7a1d'
            ns.add_usg_var_record(
                mapping_import_uuid, 'imported_mapping', mapping_var
            )
            ns.add_usg_var_record(
                intseq_import_uuid_1, 'imported_ints', intseq_var_1
            )
            ns.add_usg_var_record(
                intseq_import_uuid_2, 'imported_ints', intseq_var_2
            )

            actions = ActionCollections(
                std_actions={action_uuid: {node_uuid: 'out'}}
            )
            build_action_usage(node, ns, actions.std_actions, action_uuid, cfg)

            usg_var = ns.get_usg_var_record(node_uuid).variable

            self.assertIsInstance(usg_var, UsageVariable)
            self.assertEqual(usg_var.var_type, 'artifact')
            self.assertEqual(usg_var.name, 'out_0')

            rendered = cfg.use.render()
            self.assertIn('from qiime2 import Metadata', rendered)
            self.assertIn('.view(Metadata)', rendered)
            self.assertIn(f'.{action}(', rendered)


class BibContentTests(unittest.TestCase):
    def test_contents(self):
        series_21 = {
            'year': ' 2010 ',
            'title': ' Data Structures for Statistical Computing in Python ',
            'pages': ' 51 -- 56 ',
            'editor': ' Stéfan van der Walt and Jarrod Millman ',
            'booktitle': ' Proceedings of the 9th Python in Science Conferen',
            'author': ' Wes McKinney ',
            'ENTRYTYPE': 'inproceedings',
            'ID': 'view|types:2021.2.0|pandas.core.series:Series|0'}

        df_20 = {
            'year': ' 2010 ',
            'title': ' Data Structures for Statistical Computing in Python ',
            'pages': ' 51 -- 56 ',
            'editor': ' Stéfan van der Walt and Jarrod Millman ',
            'booktitle': ' Proceedings of the 9th Python in Science Conferen',
            'author': ' Wes McKinney ',
            'ENTRYTYPE': 'inproceedings',
            'ID': 'view|types:2020.2.0|pandas.core.frame:DataFrame|0'}

        self.assertEqual(BibContent(series_21), BibContent(df_20))
        self.assertEqual(hash(BibContent(series_21)), hash(BibContent(df_20)))

        # Set membership because these objects are equal and hash-equal
        self.assertIn(BibContent(series_21), {BibContent(df_20)})


class CitationsTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir
        cls.pm = PluginManager()

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_dedupe_citations(self):
        fn = os.path.join(self.das.datadir, 'dupes.bib')
        with open(fn) as bibtex_file:
            bib_db = bp.load(bibtex_file)
        deduped = dedupe_citations(bib_db.entries)
        # Dedupe by DOI will preserve only one of the biom.table entries
        # Dedupe by contents should preserve only one of the pandas entries
        self.assertEqual(len(deduped), 3)
        # Confirm each paper is present. The len assertion ensures one-to-one
        lower_keys = [entry['ID'].lower() for entry in deduped]
        self.assertTrue(any('framework' in key for key in lower_keys))
        self.assertTrue(any('biom' in key for key in lower_keys))
        self.assertTrue(any('pandas' in key for key in lower_keys))

    def test_dedupe_pandas(self):
        """
        No match on ID, framework, or DOI, but matching content should
        deduplicate these
        """
        series_21 = {
            'year': ' 2010 ',
            'title': ' Data Structures for Statistical Computing in Python ',
            'pages': ' 51 -- 56 ',
            'editor': ' Stéfan van der Walt and Jarrod Millman ',
            'booktitle': ' Proceedings of the 9th Python in Science Conferen',
            'author': ' Wes McKinney ',
            'ENTRYTYPE': 'inproceedings',
            'ID': 'view|types:2021.2.0|pandas.core.series:Series|0'
        }

        df_20 = {
            'year': ' 2010 ',
            'title': ' Data Structures for Statistical Computing in Python ',
            'pages': ' 51 -- 56 ',
            'editor': ' Stéfan van der Walt and Jarrod Millman ',
            'booktitle': ' Proceedings of the 9th Python in Science Conferen',
            'author': ' Wes McKinney ',
            'ENTRYTYPE': 'inproceedings',
            'ID': 'view|types:2020.2.0|pandas.core.frame:DataFrame|0'
        }

        deduped = dedupe_citations([series_21, df_20])
        self.assertEqual(len(deduped), 1)

    def test_dedupe_silva(self):
        """
        These similar publications should not be deduped by content filter
        """
        s0 = {
            'year': '2007',
            'volume': '35',
            'title': 'SILVA: a comprehensive online resource for quality '
                     'checked and aligned ribosomal RNA sequence data '
                     'compatible with ARB',
            'pages': '7188-7196',
            'number': '21',
            'journal': 'Nucleic Acids Res',
            'author': 'Pruesse, Elmar and Quast, Christian and Knittel, Katrin'
                      ' and Fuchs, Bernhard M and Ludwig, Wolfgang and Peplies'
                      ', Jorg and Glockner, Frank Oliver',
            'ENTRYTYPE': 'article',
            'ID': 'action|rescript:2020.6.0+3.g772294c|'
                  'method:parse_silva_taxonomy|0'
        }
        s1 = {
            'year': '2013',
            'volume': '41',
            'title': 'The SILVA ribosomal RNA gene database project: '
                     'improved data processing and web-based tools',
            'publisher': 'Oxford University Press',
            'pages': 'D590-6',
            'number': 'Database issue',
            'journal': 'Nucleic Acids Res',
            'author': 'Quast, Christian and Pruesse, Elmar and Yilmaz, Pelin '
                      'and Gerken, Jan and Schweer, Timmy and Yarza, Pablo and'
                      ' Peplies, Jorg and Glockner, Frank Oliver',
            'ENTRYTYPE': 'article',
            'ID': 'action|rescript:2020.6.0+3.g772294c|'
                  'method:parse_silva_taxonomy|1'
        }
        deduped = dedupe_citations([s0, s1])
        self.assertEqual(len(deduped), 2)

    def test_collect_citations_no_dedupe(self):
        dag = self.das.concated_ints_v6.dag
        exp_keys = {
            'framework|qiime2:2023.5.1|0',
            'action|dummy-plugin:0.0.0-dev|method:concatenate_ints|0',
            'plugin|dummy-plugin:0.0.0-dev|0',
            'plugin|dummy-plugin:0.0.0-dev|1',
            'view|dummy-plugin:0.0.0-dev|IntSequenceDirectoryFormat|0',
            'transformer|dummy-plugin:0.0.0-dev|'
            'builtins:list->IntSequenceDirectoryFormat|0',
            'transformer|dummy-plugin:0.0.0-dev|'
            'builtins:list->IntSequenceV2DirectoryFormat|0',
            'transformer|dummy-plugin:0.0.0-dev|'
            'builtins:list->IntSequenceV2DirectoryFormat|1',
            'transformer|dummy-plugin:0.0.0-dev|'
            'builtins:list->IntSequenceV2DirectoryFormat|2',
            'transformer|dummy-plugin:0.0.0-dev|'
            'builtins:list->IntSequenceV2DirectoryFormat|3',
            'transformer|dummy-plugin:0.0.0-dev|'
            'builtins:list->IntSequenceV2DirectoryFormat|4',
            'transformer|dummy-plugin:0.0.0-dev|'
            'builtins:list->IntSequenceV2DirectoryFormat|5',
            'transformer|dummy-plugin:0.0.0-dev|'
            'builtins:list->IntSequenceV2DirectoryFormat|6',
            'transformer|dummy-plugin:0.0.0-dev|'
            'builtins:list->IntSequenceV2DirectoryFormat|7',
            'transformer|dummy-plugin:0.0.0-dev|'
            'builtins:list->IntSequenceV2DirectoryFormat|8',
        }
        citations = collect_citations(dag, deduplicate=False)
        keys = set(citations.entries_dict.keys())
        self.assertEqual(len(keys), len(exp_keys))
        self.assertEqual(keys, exp_keys)

    def test_collect_citations_dedupe(self):
        dag = self.das.concated_ints_v6.dag
        exp_keys = {
            'framework|qiime2:2023.5.1|0',
            'action|dummy-plugin:0.0.0-dev|method:concatenate_ints|0',
            'plugin|dummy-plugin:0.0.0-dev|0',
            'plugin|dummy-plugin:0.0.0-dev|1',
            'view|dummy-plugin:0.0.0-dev|IntSequenceDirectoryFormat|0',
            'transformer|dummy-plugin:0.0.0-dev|'
            'builtins:list->IntSequenceDirectoryFormat|0',
            'transformer|dummy-plugin:0.0.0-dev|'
            'builtins:list->IntSequenceV2DirectoryFormat|4',
            'transformer|dummy-plugin:0.0.0-dev|'
            'builtins:list->IntSequenceV2DirectoryFormat|5',
            'transformer|dummy-plugin:0.0.0-dev|'
            'builtins:list->IntSequenceV2DirectoryFormat|6',
            'transformer|dummy-plugin:0.0.0-dev|'
            'builtins:list->IntSequenceV2DirectoryFormat|8'
        }

        citations = collect_citations(dag, deduplicate=True)
        print(citations.entries_dict.keys())
        keys = set(citations.entries_dict.keys())
        self.assertEqual(len(keys), len(exp_keys))
        self.assertEqual(keys, exp_keys)

    def test_collect_citations_no_prov(self):
        dag = self.das.table_v0.dag

        exp_keys = set()
        citations = collect_citations(dag)
        keys = set(citations.entries_dict.keys())
        self.assertEqual(len(keys), 0)
        self.assertEqual(keys, exp_keys)

    def test_replay_citations(self):
        dag = self.das.concated_ints_v6.dag
        exp_keys = {
            'framework|qiime2:2023.5.1|0',
            'action|dummy-plugin:0.0.0-dev|method:concatenate_ints|0',
            'plugin|dummy-plugin:0.0.0-dev|0',
            'plugin|dummy-plugin:0.0.0-dev|1',
            'view|dummy-plugin:0.0.0-dev|IntSequenceDirectoryFormat|0',
        }

        with tempfile.TemporaryDirectory() as tempdir:
            out_fp = os.path.join(tempdir, 'citations.bib')
            replay_citations(dag, out_fp)
            with open(out_fp, 'r') as fp:
                written = fp.read()
                for key in exp_keys:
                    self.assertIn(key, written)

    def test_replay_citations_no_prov(self):
        dag = self.das.table_v0.dag
        exp = "No citations were registered"
        with tempfile.TemporaryDirectory() as tempdir:
            out_fp = os.path.join(tempdir, 'citations.bib')
            replay_citations(dag, out_fp)
            with open(out_fp, 'r') as fp:
                written = fp.read()
                self.assertIn(exp, written)
